import logging
import os
import sys
from typing import Optional
from .dist_utils import get_rank
from termcolor import colored

logger_initialized = {}


class _ColorfulFormatter(logging.Formatter):
    def formatMessage(self, record):
        log = super(_ColorfulFormatter, self).formatMessage(record)
        if record.levelno == logging.DEBUG:
            prefix = colored("DEBUG", "magenta")
        elif record.levelno == logging.WARNING:
            prefix = colored("WARNING", "red", attrs=["blink"])
        elif record.levelno == logging.ERROR or record.levelno == logging.CRITICAL:
            prefix = colored("ERROR", "red", attrs=["blink", "underline"])
        else:
            return log
        return prefix + " " + log


def setup_logger(
    name: Optional[str] = None,
    output: Optional[str] = None,
    console_log_level: int = logging.INFO,
    file_log_level: int = logging.INFO,
    color: bool = False,
) -> logging.Logger:
    """
    初始化 logger

    如果logger没被初始化, 这个函数会使用1~2个handlers来初始化logger。否则这个已经初始化过的logger会直接被返回。
    在初始化时只有主程序的logger会添加handlers，以:class:`StreamHandler`的形式添加。
    如果output被赋值, :class:`FileHandler` 也会被添加

    这里是常用的用法. 我们假设文件结构如下::

        project
        ├── module1
        └── module2

    - Only setup the parent logger (``project``), then all children loggers
      (``project.module1`` and ``project.module2``) will use the handlers of the parent logger.

    Example::

        >>> setup_logger(name="project")
        >>> logging.getLogger("project.module1")
        >>> logging.getLogger("project.module2")

    - Only setup the root logger, then all loggers will use the handlers of the root logger.

    Example::

        >>> setup_logger()
        >>> logging.getLogger(name="project")
        >>> logging.getLogger(name="project.module1")
        >>> logging.getLogger(name="project.module2")

    - Setup all loggers, each logger uses independent handlers.

    Example::

        >>> setup_logger(name="project")
        >>> setup_logger(name="project.module1")
        >>> setup_logger(name="project.module2")

    Parameters
    ----------
    name : str, default None
        Logger 名字。
    output : str, default None
        一个保存log的文件名或目录名
        * None: 不会保存log到文件
        * 后缀带.txt或.log : 将其设置成文件名
        * other : 文件名变为 output/log.txt
    console_log_level : int, default logging.INFO
        logger输出到控制台/终端的等级
    file_log_level : int, default logging.INFO
        logger输出到文件的等级
    color : bool, default False
        如果是True，logger将会有颜色

    Returns
    -------
        logging.Logger: 一个已初始化的logger
    """
    if name in logger_initialized:
        return logger_initialized[name]

    # get root logger if name is None
    logger = logging.getLogger(name)
    logger.setLevel(console_log_level)
    # the messages of this logger are not propagated to its parent
    logger.propagate = False

    plain_formatter = logging.Formatter(
        "[%(asctime)s %(name)s %(levelname)s]: %(message)s", datefmt="%m/%d %H:%M:%S"
    )

    # stdout and file logging: master only
    if get_rank() == 0:
        ch = logging.StreamHandler(stream=sys.stdout)
        ch.setLevel(console_log_level)
        if color:
            formatter = _ColorfulFormatter(
                colored("[%(asctime)s %(name)s]: ", "green") + "%(message)s",
                datefmt="%m/%d %H:%M:%S",
            )
        else:
            formatter = plain_formatter
        ch.setFormatter(formatter)
        logger.addHandler(ch)

        if output is not None:
            if output.endswith(".txt") or output.endswith(".log"):
                filename = output
            else:
                filename = os.path.join(output, "log.txt")
            # If a single file name is passed as argument, os.path.dirname() will return an empty
            # string. For example, os.path.dirname("log.txt") == "". This will cause an error
            # in os.makedirs(). So we need to wrap `filename` with os.path.abspath().
            os.makedirs(os.path.dirname(os.path.abspath(filename)), exist_ok=True)

            fh = logging.FileHandler(filename)
            fh.setLevel(file_log_level)
            fh.setFormatter(plain_formatter)
            logger.addHandler(fh)

    logger_initialized[name] = logger
    return logger



import logging
import os
import os.path as osp
import time
import weakref
from typing import Any, Dict, List, Optional, Tuple
import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from torch.cuda.amp import GradScaler, autocast
from torch.nn.parallel import DataParallel, DistributedDataParallel
from torch.nn.utils import clip_grad_norm_
from torch.utils.data import DataLoader
from .hooks import CheckpointerHook, HookBase, LoggerHook
from .utils import setup_logger, ProgressBar, HistoryBuffer, EMA
from .utils.misc import get_workspace
from .utils.dist_utils import get_rank
from .lr_scheduler import LRWarmupScheduler
from ._get_logger import logger


class Trainer:
    """
    一个基于epoch的通用训练框架（目前只支持单gpu运行）, 包含：
    1. 计算从dataloader中计算loss
    2. 计算梯度
    3. 用optimizer更新参数
    4. 调整学习率
    如果想完成更复杂的功能, 也可以继承该类编写子类, 重写里面的`train_one_iter`等方法
    以下是代码示例
    .. code-block:: python
        model = ...          # 初始化你的模型
        optimizer = ...      # 你的优化器
        lr_scheduler = ...   # 初始化你的调节器
        data_loader = ...    # 初始化你的数据生成器
        # 训练100轮
        trainer = Trainer(model, optimizer, lr_scheduler, data_loader, max_epochs=100)
        trainer.train()

    Parameters
    ---------
    model : torch.nn.Module, 训练模型, 训练时的输出只能是以下三种:
        * torch.Tensor, 对于这种输出是模型backward用的loss, 在torch-frame框架中被称为total_loss
        * dict, 里面是模型的各路分支的loss，需要是标量, Trainer会自动将其求和得到total_loss, 再做backward
        * Tuple[Union[dict, torch.Tensor], dict]. 前两种的混合体, 元组第一个输出是前面两种的任意一种;
          第二个输出是非需要backward类的, Trainer不会把这个dict汇总到total_loss上
    optimizer : torch.optim.Optimizer, 优化器
    lr_scheduler : optim.lr_scheduler._LRScheduler, 学习率调节器
    data_loader : torch.utils.data.DataLoader, 数据生成器
    max_epochs : int, 训练的总轮数
    work_dir : str, 保存模型和日志的根目录地址
    clip_grad_norm : float, default 0.0
        梯度裁剪的设置, 如果置为小于等于0, 则不作梯度裁剪
    enable_amp : bool, 使用混合精度
    warmup_method : str, default None
        warmup的类型, 包含以下四种取值
        * constant
        * linear
        * exp
        * None : 不使用warmup
    warmup_iters : int, default 1000, warmup最后的iter数
    warmup_factor : float, default 0.001
        warmup初始学习率 = warmup_factor * initial_lr
    hooks : List[HookBase], default None.
        hooks, 保存模型、输出评估指标、loss等用
    use_ema : bool, default False. 是否使用EMA技术
    ema_decay: float = 0.9999. EMA模型衰减系数
    create_new_dir : Optional[str], default time
        存在同名目录时以何种策略创建目录
        * None, 直接使用同名目录
        * `time_s`, 如果已经存在同名目录, 则以时间(精确到秒)为后缀创建新目录
        * `time_m`, 如果已经存在同名目录, 则以时间(精确到分)为后缀创建新目录
        * `time_h`, 如果已经存在同名目录, 则以时间(精确到小时)为后缀创建新目录
        * `time_d`, 如果已经存在同名目录, 则以时间(精确到日)为后缀创建新目录
        * `count`, 如果已经存在同名目录, 则以序号为后缀创建新目录
    """

    def __init__(
            self,
            model: nn.Module,
            optimizer: optim.Optimizer,
            lr_scheduler: optim.lr_scheduler._LRScheduler,
            data_loader: DataLoader,
            max_epochs: int,
            work_dir: str = "work_dir",
            clip_grad_norm: float = 0.0,
            enable_amp=False,
            warmup_method: Optional[str] = None,
            warmup_iters: int = 1000,
            warmup_factor: float = 0.001,
            hooks: Optional[List[HookBase]] = None,
            use_ema: bool = False,
            ema_decay: float = 0.9999,
            create_new_dir: Optional[str] = "time_s"
    ):
        logger.setLevel(logging.INFO)

        self.work_dir = get_workspace(work_dir, create_new_dir)

        self.model = model
        if use_ema:
            self.model_ema = EMA(self.model_or_module, ema_decay)
        else:
            self.model_ema = None
        self.optimizer = optimizer
        # convert epoch-based scheduler to iteration-based scheduler
        self.lr_scheduler = LRWarmupScheduler(
            lr_scheduler, len(data_loader), warmup_method, warmup_iters, warmup_factor
        )
        self.data_loader = data_loader
        self.metric_storage = MetricStorage()

        # counters
        self.inner_iter: int = -1  # [0, epoch_len - 1]
        self.epoch: int = -1  # [0, max_epochs - 1]
        self.start_epoch = 0  # [0, max_epochs - 1]
        self.max_epochs = max_epochs

        self._hooks: List[HookBase] = []
        self._clip_grad_norm = clip_grad_norm
        if not torch.cuda.is_available():
            enable_amp = False
            logger.warning("torch环境无法使用cuda, AMP无法使用")
        self._enable_amp = enable_amp

        if self._enable_amp:
            logger.info("自动混合精度 (AMP) 训练")
            self._grad_scaler = GradScaler()

        self.rank = get_rank()
        if hooks is None:
            self.register_hooks(self._build_default_hooks())
        elif self.rank != 0:
            logger.warning(f"{self.rank}号卡也被传入了hook, 将被自动忽略")
        else:
            self.register_hooks(hooks)

    @property
    def lr(self) -> float:
        return self.optimizer.param_groups[0]["lr"]

    @property
    def epoch_len(self) -> int:
        return len(self.data_loader)

    @property
    def max_iters(self) -> int:
        return self.max_epochs * self.epoch_len

    @property
    def cur_iter(self) -> int:
        """返回当前iter数, 范围在 [0, max_iters - 1]."""
        return self.epoch * self.epoch_len + self.inner_iter

    @property
    def start_iter(self) -> int:
        """从哪一个iter开始. 最小的值是0."""
        return self.start_epoch * self.epoch_len

    @property
    def ckpt_dir(self) -> str:
        return osp.join(self.work_dir, "checkpoints")

    @property
    def tb_log_dir(self) -> str:
        return osp.join(self.work_dir, "tb_logs")

    @property
    def log_file(self) -> str:
        return osp.join(self.work_dir, "log.txt")

    @property
    def model_or_module(self) -> nn.Module:
        if isinstance(self.model, (DistributedDataParallel, DataParallel)):
            return self.model.module
        return self.model

    @property
    def model_evaluate(self) -> nn.Module:
        """评估用的模型"""
        if self.model_ema:
            return self.model_ema.model
        return self.model_or_module

    @property
    def registered_hook_names(self) -> List[str]:
        """注册的所有hook名字"""
        return [h.__class__.__name__ for h in self._hooks]

    def log(self, *args, **kwargs) -> None:
        """更新评估指标"""
        self.metric_storage.update(*args, **kwargs)

    def _prepare_for_training(self,
                              console_log_level: int = 2,
                              file_log_level: int = 2) -> None:
        """
        训练前的配置工作
        Parameters
        ----------
        console_log_level : int, default 2
             输出到屏幕的log等级, 可选范围是0-5, 它们对应的关系分别为：
             * 5: FATAL
             * 4: ERROR
             * 3: WARNING
             * 2: INFO
             * 1: DEBUG
             * 0: NOTSET
        file_log_level : int, default 2
             输出到文件里的log等级, 其他方面同console_log_level参数
        """
        # setup the root logger of the `cpu` library to show
        # the log messages generated from this library
        assert console_log_level in (0, 1, 2, 3, 4, 5), f"console_log_level必须在0~5之间而不是{console_log_level}"
        assert file_log_level in (0, 1, 2, 3, 4, 5), f"file_log_level必须在0~5之间而不是{file_log_level}"
        console_log_level *= 10
        file_log_level *= 10
        setup_logger("torch_frame", output=self.log_file,
                     console_log_level=console_log_level, file_log_level=file_log_level)

        os.makedirs(self.ckpt_dir, exist_ok=True)
        if self.start_epoch == 0:
            split_line = "-" * 50
            logger.info(
                f"\n{split_line}\n"
                f"Work directory: {self.work_dir}\n"
                f"{split_line}"
            )

    def register_hooks(self, hooks: List[Optional[HookBase]]) -> None:
        """
        Trainer运行时调用hook
        hook执行时根据它们注册的顺序来进行

        Parameters
        ---------
        hooks : list[HookBase]
        """
        hooks = [h for h in hooks if h is not None]
        for h in hooks:
            assert isinstance(h, HookBase)
            # To avoid circular reference, hooks and trainer cannot own each other. This normally
            # does not matter, but will cause memory leak if the involved objects contain __del__.
            # See http://engineering.hearsaysocial.com/2013/06/16/circular-references-in-python/
            h.trainer = weakref.proxy(self)
            # We always keep :class:`LoggerHook` as the last hook to avoid losing any records
            # that should have been logged. The order of other hooks remains the same.
            if self._hooks and isinstance(self._hooks[-1], LoggerHook):
                self._hooks.insert(len(self._hooks) - 1, h)
            else:
                self._hooks.append(h)
        logger.warning(f"Registered default hooks: {self.registered_hook_names}")

    def _call_hooks(self, stage: str) -> None:
        for h in self._hooks:
            getattr(h, stage)()

    def _build_default_hooks(self) -> List[HookBase]:
        return [
            CheckpointerHook(),
            LoggerHook(tb_log_dir=self.tb_log_dir),
        ]

    def _update_iter_metrics(self, loss_dict: Dict[str, torch.Tensor], data_time: float,
                             iter_time: float, lr: float) -> None:
        """
        每个iter评估的log
        Parameters
        ----------
        loss_dict : dict, losses的标量字典
        data_time : float, dataloader的一个iter耗时
        iter_time : float, 一个iter全部耗时
        lr : float, 该iter的学习率
        """
        self.log(self.cur_iter, data_time=data_time, iter_time=iter_time)
        self.log(self.cur_iter, lr=lr, smooth=False)

        loss_value = sum(loss_dict.values())
        if not np.isfinite(loss_value):
            raise FloatingPointError(
                f"Loss became infinite or NaN at epoch={self.epoch}! loss_dict = {loss_dict}."
            )

        self.log(self.cur_iter, **loss_dict)

    def train_one_iter(self, batch) -> dict:
        """
        包含了训练的一个iter的全部操作

        .. Note::
            标准的学习率调节器是基于epoch的, 但torch_frame框架是基于iter的, 所以它在每次iter之后都会调用
        """
        #####################
        # 1. 计算loss #
        #####################
        if self._enable_amp:
            with autocast():
                loss_info = self.model(batch)
        else:
            loss_info = self.model(batch)
        if isinstance(loss_info, torch.Tensor):
            losses = loss_info
            loss_info = {"total_loss": loss_info}
        elif isinstance(loss_info, tuple):
            assert len(loss_info) == 2, "loss_info需要是一个二元组，第一个是需要反向传播的，第二个是其他参考指标"
            backward_params, metric_params = loss_info
            assert isinstance(metric_params, dict), "loss_info的第二个值需要是dict类型"
            if isinstance(backward_params, torch.Tensor):
                losses = backward_params
                metric_params["total_loss"] = backward_params
                loss_info = metric_params
            elif isinstance(backward_params, dict):
                losses = sum(backward_params.values())
                backward_params["total_loss"] = losses
                backward_params.update(metric_params)
                loss_info = backward_params
        else:
            assert "total_loss" not in loss_info, "当model返回是一个dict的时候不可以传出包含"
            losses = sum(loss_info.values())
            loss_info["total_loss"] = losses

        ##########################
        # 2. 计算梯度 #
        ##########################
        self.optimizer.zero_grad()
        if self._enable_amp:
            self._grad_scaler.scale(losses).backward()
        else:
            losses.backward()
        if self._clip_grad_norm > 0:
            if self._enable_amp:
                self._grad_scaler.unscale_(self.optimizer)
            clip_grad_norm_(self.model.parameters(), self._clip_grad_norm)

        ##############################
        # 3. 更新模型参数 #
        ##############################
        if self._enable_amp:
            self._grad_scaler.step(self.optimizer)
            self._grad_scaler.update()
        else:
            self.optimizer.step()
        if self.model_ema:
            self.model_ema.update(self.model)

        ###########################
        # 4. 调整学习率 #
        ###########################
        self.lr_scheduler.step()

        show_info = {k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else v for k, v in loss_info.items()}
        show_info = dict(sorted(show_info.items(), key=lambda x: x[0] != "total_loss"))  # 保证total_loss在最后一位
        self.pbar.set_postfix(show_info)
        return show_info

    def _train_one_epoch(self) -> None:
        """执行模型一个epoch的全部操作"""
        self.model.train()
        self.pbar = ProgressBar(total=self.epoch_len, desc=f"epoch={self.epoch}", ascii=True)

        start_time_data = time.perf_counter()
        for self.inner_iter, batch in enumerate(self.data_loader):
            start_time_iter = time.perf_counter()
            data_time = start_time_iter - start_time_data
            self._call_hooks("before_iter")
            show_info = self.train_one_iter(batch)
            self._call_hooks("after_iter")
            self._update_iter_metrics(show_info, data_time, time.perf_counter() - start_time_data, self.lr)
            self.pbar.update(1)
            start_time_data = time.perf_counter()
        self.pbar.close()
        del self.pbar

    def train(self,
              console_log_level: int = 2,
              file_log_level: int = 2) -> None:
        """
        训练入口

        Parameters
        ----------
        console_log_level : int, default 2.
            输出到屏幕的log等级, 可选范围是0-5, 它们对应的关系分别为：
            * 5: FATAL
            * 4: ERROR
            * 3: WARNING
            * 2: INFO
            * 1: DEBUG
            * 0: NOTSET
        file_log_level : int, default 2.
            输出到文件里的log等级, 其他方面同console_log_level参数
        """
        logger.info(f"开始从第{self.start_epoch}epoch训练")
        self._prepare_for_training(console_log_level, file_log_level)
        self._call_hooks("before_train")
        for self.epoch in range(self.start_epoch, self.max_epochs):
            self._call_hooks("before_epoch")
            self._train_one_epoch()
            self._call_hooks("after_epoch")
        self._call_hooks("after_train")

    def save_checkpoint(self, file_name: str, save_single_model: bool = True,
                        print_info: bool = False) -> None:
        """
        保存参数, 包含:

        * epoch : 当前轮数
        * model : 当前模型参数
        * optimizer : 当前优化器
        * lr_scheduler : 当前调节器
        * metric_storage : 评估指标
        * hooks(非必须) : 一些中间量
        * grad_scaler(非必须) : 混合精度的参数

        Parameters
        ----------
        file_name : str, 保存文件名
        save_single_model : bool, default True.
            * True, 会保存模型参数本身
            * False, 会保存包含模型、hook及优化器等权重
        print_info : bool, default True. 如果是True, 则输出保存模型的提示信息
        """

        data = {
            "epoch": self.epoch,
            "model": self.model_or_module.state_dict(),
            "optimizer": self.optimizer.state_dict(),
            "lr_scheduler": self.lr_scheduler.state_dict(),
            "metric_storage": self.metric_storage,
            "work_dir": self.work_dir,
        }
        hook_states = {h.class_name: h.state_dict() for h in self._hooks if h.checkpointable}
        if hook_states:
            data["hooks"] = hook_states
        if hasattr(self, "_enable_amp") and self._enable_amp:
            data["grad_scaler"] = self._grad_scaler.state_dict()

        file_path = osp.join(self.ckpt_dir, file_name)
        if print_info:
            logger.info(f"Saving checkpoint to {file_path}")
        if save_single_model:
            torch.save(self.model_evaluate.state_dict(), file_path)
        else:
            torch.save(data, file_path)

    def load_checkpoint(self, path: str = None, checkpoint: Dict[str, Any] = None):
        """
        加载参数

        Parameters
        ----------
        path : str, default None. checkpoint的地址
        checkpoint : dict, default None.
            如果path非空, 优先使用path的数据, 否则直接加载checkpoint的数据。
            直接加载的时候，将只加载模型，而不带各种状态
        """
        assert checkpoint is None or path is None
        if path is None:
            incompatible = self.model_or_module.load_state_dict(checkpoint, strict=False)
            if incompatible.missing_keys:
                logger.warning("Encounter missing keys when loading model weights:\n"
                               f"{incompatible.missing_keys}")
            if incompatible.unexpected_keys:
                logger.warning("Encounter unexpected keys when loading model weights:\n"
                               f"{incompatible.unexpected_keys}")
            logger.info("只加载模型本身...")
            return

        checkpoint = torch.load(path, map_location="cpu")

        # 1. 加载 epoch
        self.start_epoch = checkpoint["epoch"] + 1

        # 2. 加载 metric_storage
        self.metric_storage = checkpoint["metric_storage"]

        # 3. 加载 optimizer
        self.optimizer.load_state_dict(checkpoint["optimizer"])

        # 4. 加载 lr_scheduler
        self.lr_scheduler.load_state_dict(checkpoint["lr_scheduler"])

        # 5. 加载 grad scaler
        consistent_amp = not (self._enable_amp ^ ("grad_scaler" in checkpoint))
        assert consistent_amp, "Found inconsistent AMP training setting when loading checkpoint."
        if hasattr(self, "_enable_amp") and self._enable_amp:
            self._grad_scaler.load_state_dict(checkpoint["grad_scaler"])

        # 6. 加载 模型
        incompatible = self.model_or_module.load_state_dict(checkpoint["model"], strict=False)
        if incompatible.missing_keys:
            logger.warning("Encounter missing keys when loading model weights:\n"
                           f"{incompatible.missing_keys}")
        if incompatible.unexpected_keys:
            logger.warning("Encounter unexpected keys when loading model weights:\n"
                           f"{incompatible.unexpected_keys}")

        # 7. 加载 hooks
        hook_states = checkpoint.get("hooks", {})
        hook_names = [h.class_name for h in self._hooks if h.checkpointable]
        missing_keys = [name for name in hook_names if name not in hook_states]
        unexpected_keys = [key for key in hook_states if key not in hook_names]
        if missing_keys:
            logger.warning(f"Encounter missing keys when loading hook state dict:\n{missing_keys}")
        if unexpected_keys:
            logger.warning(f"Encounter unexpected keys when loading hook state dict:\n{unexpected_keys}")

        for key, value in hook_states.items():
            for h in self._hooks:
                if h.class_name == key and h.checkpointable:
                    h.load_state_dict(value)
                    break

        # 8. 加载保存目录
        self.work_dir = checkpoint["work_dir"]

        # 9. 加载ema 模型
        if self.model_ema:
            lambda_decay = self.model_ema.decay
            self.model_ema = EMA(self.model_or_module, updates=self.cur_iter)
            self.model_ema.decay = lambda_decay

        if path:
            logger.info(f"加载模型{path}成功")


class MetricStorage(dict):
    """The class stores the values of multiple metrics (some of them may be noisy, e.g., loss,
    batch time) in training process, and provides access to the smoothed values for better logging.

    The class is designed for automatic tensorboard logging. User should specify the ``smooth``
    when calling :meth:`update`, in order to we can determine which metrics should be
    smoothed when performing tensorboard logging.

    Example::

        >>> metric_storage = MetricStorage()
        >>> metric_storage.update(iter=0, loss=0.2)
        >>> metric_storage.update(iter=0, lr=0.01, smooth=False)
        >>> metric_storage.update(iter=1, loss=0.1)
        >>> metric_storage.update(iter=1, lr=0.001, smooth=False)
        >>> # loss will be smoothed, but lr will not
        >>> metric_storage.values_maybe_smooth
        {"loss": (1, 0.15), "lr": (1, 0.001)}
        >>> # like dict, can be indexed by string
        >>> metric_storage["loss"].avg
        0.15
    """

    def __init__(self, default_win_size: int = 20) -> None:
        self._default_win_size = default_win_size
        self._history: Dict[str, HistoryBuffer] = self
        self._smooth: Dict[str, bool] = {}
        self._latest_iter: Dict[str, int] = {}

    def update(self, iter: Optional[int] = None, smooth: bool = True, window_size: int = None, **kwargs) -> None:
        """Add new scalar values of multiple metrics produced at a certain iteration.

        Args:
            iter (int): The iteration in which these values are produced.
                If None, use the built-in counter starting from 0.
            smooth (bool): If True, return the smoothed values of these metrics when
                calling :meth:`values_maybe_smooth`. Otherwise, return the latest values.
                The same metric must have the same ``smooth`` in different calls to :meth:`update`.
            window_size : int
        """
        for key, value in kwargs.items():
            if key in self._smooth:
                assert self._smooth[key] == smooth
            else:
                self._smooth[key] = smooth
                self._history[key] = HistoryBuffer(window_size=window_size if window_size else self._default_win_size)
                self._latest_iter[key] = -1
            if iter is not None:
                assert iter > self._latest_iter[key], "检查total_loss是不是存在于model给出的loss_dict中"
                self._latest_iter[key] = iter
            else:
                self._latest_iter[key] += 1
            self._history[key].update(value)

    @property
    def values_maybe_smooth(self) -> Dict[str, Tuple[int, float]]:
        """Return the smoothed values or the latest values of multiple metrics.
        The specific behavior depends on the ``smooth`` when updating metrics.

        Returns:
            dict[str -> (int, float)]: Mapping from metric name to its
                (the latest iteration, the avg/latest value) pair.
        """
        return {
            key: (self._latest_iter[key], his_buf.avg if self._smooth[key] else his_buf.latest)
            for key, his_buf in self._history.items()
        }



import tqdm
from .dist_utils import is_dist_avail_and_initialized, get_rank


class ProgressBar:
    """tqdm基础上包了一层，主要兼容ddp的多卡"""
    def __init__(self, *args, **kwargs):
        if not is_dist_avail_and_initialized() or get_rank() == 0:
            self.pbar = tqdm.tqdm(*args, **kwargs)
        else:
            self.pbar = None

    def update(self, *args, **kwargs):
        if self.pbar is not None:
            self.pbar.update(*args, **kwargs)

    def close(self):
        if self.pbar is not None:
            self.pbar.close()

    def set_postfix(self, *args, **kwargs):
        if self.pbar is not None:
            self.pbar.set_postfix(*args, **kwargs)

    def set_description(self, *args, **kwargs):
        if self.pbar is not None:
            self.pbar.set_description(*args, **kwargs)

    def set_postfix_str(self, *args, **kwargs):
        if self.pbar is not None:
            self.pbar.set_postfix_str(*args, **kwargs)

    def set_description_str(self, *args, **kwargs):
        if self.pbar is not None:
            self.pbar.set_description_str(*args, **kwargs)


import numpy as np


__all__ = [
    "RepeatDataset",
]


# 源自https://github.com/open-mmlab/mmdetection/blob/master/mmdet/datasets/dataset_wrappers.py
class RepeatDataset:
    """A wrapper of repeated dataset.

    The length of repeated dataset will be `times` larger than the original
    dataset. This is useful when the data loading time is long but the dataset
    is small. Using RepeatDataset can reduce the data loading time between
    epochs.

    Args:
        dataset (:obj:`Dataset`): The dataset to be repeated.
        times (int): Repeat times.
    """

    def __init__(self, dataset, times):
        self.dataset = dataset
        self.times = times

        self._ori_len = len(self.dataset)

    def __getitem__(self, idx):
        return self.dataset[idx % self._ori_len]

    def __len__(self):
        """Length after repetition."""
        return self.times * self._ori_len


"""
包含ddp相关函数
"""

import torch.distributed as dist


def is_dist_avail_and_initialized():
    if not dist.is_available():
        return False
    if not dist.is_initialized():
        return False
    return True


def get_world_size():
    if not is_dist_avail_and_initialized():
        return 1
    return dist.get_world_size()


def get_rank():
    if not is_dist_avail_and_initialized():
        return 0
    return dist.get_rank()


"""
该脚本包含主流评估方法, 区别于history_buffer, 没有窗口大小的概念, 换言之不需要做平滑之类的操作, 而求所有数据的精准指标
"""
from abc import abstractmethod
from functools import partial
from typing import Dict, Optional, Callable

import numpy as np

from ..vision import det_postprocess, eval_map


class BaseMetric:
    """评估器基类, 继承它需要实现`update`, `evaluate`和`reset`方法"""

    @abstractmethod
    def update(self, *args, **kwargs):
        ...

    @abstractmethod
    def evaluate(self):
        ...

    @abstractmethod
    def reset(self):
        ...


class ModelMetric(BaseMetric):
    """
    模型评估器将update拆分成模型推理的inference和结果收集步骤.
    继承它需要实现`inference`, `evaluate`和`reset`方法
    """

    def update(self, model, inputs):
        result = self.inference(model, inputs)
        self.collection(result)

    @abstractmethod
    def inference(self, model, inputs):
        ...

    @abstractmethod
    def collection(self, item):
        ...

    @abstractmethod
    def evaluate(self):
        ...

    @abstractmethod
    def reset(self):
        ...


class ObjDetMAPMetric(ModelMetric):
    """
    目标检测中的AP评估器
    """

    def __init__(self,
                 class2idx: Dict[str, int],
                 is_xyxy: bool = False,
                 ovthresh: float = 0.5,
                 postprocess: Optional[Callable] = None,
                 confthre: float = 0.7,
                 nmsthre: float = 0.45,
                 ):
        """
        Parameters
        ----------
        class2idx : Dict[str, int]. 类别到索引的映射, 不包含背景类
        is_xyxy: bool, default False. gt框类型
            * False. gt目标框是[cx, cy, w, h]类型
            * True. gt目标框是[x1, y1, x2, y2]类型
        ovthresh : float, default 0.5. 计算AP时，pred和gt被认为是TP（true positive）的iou阈值
        postprocess : Optional[Callable], default None. 承接模型推理后的后处理
            * None, 则默认Yolo风格的内置函数det_postprocess
            * Callable, 返回是一个list或np.ndarray. 每一行是[x1, y1, x2, y2, score, cls]
        confthre : float, default 0.7. 如果postprocess使用默认值, 则为det_postprocess的对应参数, postprocess非空不生效!
        nmsthre : float, default 0.45. 如果postprocess使用默认值, 则为det_postprocess的对应参数, postprocess非空不生效!
        """
        self.class2idx = class2idx
        self.is_xyxy = is_xyxy
        self.ovthresh = ovthresh
        self.idx2class = {cls: idx for cls, idx in class2idx.items()}
        if postprocess is not None:
            self.postprocess = postprocess
        else:
            self.postprocess = partial(det_postprocess,
                                       num_classes=len(self.class2idx),
                                       conf_thre=confthre,
                                       nms_thre=nmsthre,
                                       class_agnostic=True,
                                       merge_score=True,
                                       ret_np=True
                                       )
        self.all_gt_boxes = {cls: [] for cls in self.class2idx}
        self.all_pred_boxes = {cls: [] for cls in self.class2idx}

    def inference(self, model, inputs):
        image, gt = inputs

        outputs = model(image)
        return outputs, gt

    def collection(self, item):
        outputs, gt = item
        outputs = self.postprocess(outputs)
        for i, one_image_boxes in enumerate(outputs):
            if one_image_boxes is None:
                for boxes in self.all_pred_boxes.values():
                    boxes.append({})
                continue
            for cls_name, cls_idx in self.class2idx.items():
                mask = one_image_boxes[..., -1] == cls_idx
                if np.any(mask) == 0:
                    self.all_pred_boxes[cls_name].append({})
                else:
                    image_cls_boxes = {
                        "confidence": one_image_boxes[mask][:, -2],
                        "bboxes": one_image_boxes[mask][:, :4],
                    }
                    self.all_pred_boxes[cls_name].append(image_cls_boxes)
        for i, one_image_boxes in enumerate(gt):
            one_image_boxes = one_image_boxes[one_image_boxes.sum(-1) > 0].numpy()
            if not self.is_xyxy:
                cxcywh = one_image_boxes[:, 1:]
                one_image_boxes = np.stack([
                    one_image_boxes[:, 0],
                    cxcywh[:, 0] - cxcywh[:, 2] / 2,
                    cxcywh[:, 1] - cxcywh[:, 3] / 2,
                    cxcywh[:, 0] + cxcywh[:, 2] / 2,
                    cxcywh[:, 1] + cxcywh[:, 3] / 2,
                ], -1)
            for cls_name, cls_idx in self.class2idx.items():
                mask = one_image_boxes[..., 0] == cls_idx
                if np.any(mask) == 0:
                    self.all_gt_boxes[cls_name].append({})
                else:
                    image_cls_boxes = {
                        "labels": one_image_boxes[mask][:, 0],
                        "bboxes": one_image_boxes[mask][:, 1:],
                    }
                    self.all_gt_boxes[cls_name].append(image_cls_boxes)

    def evaluate(self):
        aps = eval_map(self.all_gt_boxes, self.all_pred_boxes, self.ovthresh)
        new_aps = {f"ap_{k}": v for k, v in aps.items()}
        new_aps["map"] = np.mean(list(new_aps.values()))
        return new_aps

    def reset(self):
        self.all_gt_boxes = {cls: [] for cls in self.class2idx}
        self.all_pred_boxes = {cls: [] for cls in self.class2idx}



import os
import pickle
import os.path as osp
from typing import Any, Dict, List, Optional
from types import LambdaType
import logging
from .hookbase import HookBase

logger = logging.getLogger(__name__)


class CheckpointerHook(HookBase):
    """
    周期性保存参数
    """

    def __init__(self, period: int = 1,
                 max_to_keep: Optional[int] = None,
                 save_metric: Optional[str] = None,
                 max_first: bool = True,
                 save_last: bool = True
                 ) -> None:
        """
        初始化
        Parameters
        ----------
        period : int, 保存checkpoints的周期
        max_to_keep : int, 保存checkpoints的数量, 更早期的checkpoints会被删除
        save_metric : int, default None.
            保存模型的指标是哪个, 需要从trainer.metric_storage选择
        max_first : bool, default True.
            用于保存模型的指标是取最大还是最小作为最优模型
        save_last : bool, default True
            是否保存最近一次的epoch的模型, 如果是True, 每轮将更新模型到latest.pth中
        """
        self._period = period
        assert max_to_keep is None or max_to_keep > 0
        self._max_to_keep = max_to_keep
        if save_metric is None:
            logger.warning("没有指定保存模型的指标，因此每period都将保存模型")
        self.save_metric = save_metric
        if max_first:
            self.cur_best = float("-inf")
            self.is_better = lambda a: a > self.cur_best
        else:
            self.cur_best = float("inf")
            self.is_better = lambda a: a < self.cur_best

        self.save_last = save_last

        self._recent_checkpoints: List[str] = []

    def after_epoch(self) -> None:
        if self.every_n_epochs(self._period) or self.is_last_epoch():
            self.save_model()

    def save_model(self):
        if self.save_last:
            self.trainer.save_checkpoint("latest.pth", False)

        # 如果当前epoch指标没有更好, 则不保存模型
        if self.save_metric is not None:
            if not self.is_better(self.trainer.metric_storage[self.save_metric]):
                return
            else:
                self.cur_best = self.trainer.metric_storage[self.save_metric].avg
                logger.info(f"{self.save_metric} update to {round(self.cur_best, 4)}")

        self.trainer.save_checkpoint("best.pth")
        if self._max_to_keep is not None and self._max_to_keep >= 1:
            epoch = self.trainer.epoch  # ranged in [0, max_epochs - 1]
            checkpoint_name = f"epoch_{epoch}.pth"
            self.trainer.save_checkpoint(checkpoint_name)
            self._recent_checkpoints.append(checkpoint_name)
            if len(self._recent_checkpoints) > self._max_to_keep:
                # delete the oldest checkpoint
                file_name = self._recent_checkpoints.pop(0)
                file_path = osp.join(self.trainer.ckpt_dir, file_name)
                if os.path.exists(file_path):
                    os.remove(file_path)

    def state_dict(self) -> Dict[str, Any]:
        state = {}
        for key, value in self.__dict__.items():
            if key == "trainer" or isinstance(value, LambdaType):
                continue
            try:
                pickle.dumps(value)
            except BaseException:
                continue
            state[key] = value
        return state

    def load_state_dict(self, state_dict: Dict[str, Any]) -> None:
        self.__dict__.update(state_dict)



from typing import Callable, Optional
from tqdm import tqdm
import torch
from torch.utils.data.dataloader import DataLoader
from .checkpoint_hook import CheckpointerHook
import numpy as np


class EvalHook(CheckpointerHook):
    """`CheckpointerHook` 的派生类, 周期性执行的评估器, 在每个epoch的最后阶段执行"""

    def __init__(self,
                 dataloader: DataLoader,
                 eval_func: Callable,
                 period: int = 1,
                 max_to_keep: Optional[int] = None,
                 save_metric: Optional[str] = None,
                 max_first: bool = True,
                 save_last: bool = True,
                 prefix: str = "eval"
                 ):
        """

        Parameters
        ----------
        dataloader : DataLoader.
            测试数据的dataloader
        eval_func : Callable.
            一个函数, 输入参数是model和batch, 返回一个评估结果的Dict[list], k对应指标名称, v是包含每个样本得分的list
        period : int, default 1.
            执行eval_func函数的周期
        max_to_keep : int, 保存checkpoints的数量, 更早期的checkpoints会被删除
        save_metric : int, default None.
            保存模型的指标是哪个, 需要从trainer.metric_storage选择
        max_first : bool, default True.
            用于保存模型的指标是取最大还是最小作为最优模型
        save_last : bool, default True
            是否保存最近一次的epoch的模型, 如果是True, 每轮将更新模型到latest.pth中
        """
        self.prefix = prefix+"_"
        super(EvalHook, self).__init__(period, max_to_keep, self.prefix + save_metric, max_first, save_last)
        self._eval_func = eval_func
        self.dataloader = dataloader

    @torch.no_grad()
    def _do_eval(self):
        tot_res = {}
        mode_model = self.trainer.model_evaluate.training
        self.trainer.model_evaluate.eval()
        with tqdm(self.dataloader, desc="eval") as pbar:
            for batch in pbar:
                res = self._eval_func(self.trainer.model_evaluate, batch)
                for k, v in res.items():
                    tot_res.setdefault(k, []).extend(v)
        self.trainer.model_evaluate.train(mode_model)
        if tot_res:
            rename_res = {self.prefix + k: np.mean(v) for k, v in tot_res.items()}
            self.log(self.trainer.epoch, **rename_res, smooth=False, window_size=1)

    def after_epoch(self):
        if self.every_n_epochs(self._period) or self.is_last_epoch():
            self._do_eval()
            self.save_model()


class EvalTotalHook(CheckpointerHook):
    """
    `CheckpointerHook` 的派生类, 周期性执行的评估器, 在每个epoch的最后阶段执行.
    与`EvalHook`评估器区别是: 一个是每个batch都去评估, 最后求每次评估结果的均值; 一个是先把batch结果存下来, 最后一起评估
    """

    def __init__(self,
                 dataloader: DataLoader,
                 eval_metric: object,
                 period: int = 1,
                 max_to_keep: Optional[int] = None,
                 save_metric: Optional[str] = None,
                 max_first: bool = True,
                 save_last: bool = True,
                 prefix: str = "eval"
                 ):
        """

        Parameters
        ----------
        dataloader : DataLoader.
            测试数据的dataloader
        eval_metric : object.
            一个对象, 需要包含`update`和`evaluate`方法. 其中:
            * `update`方法, 需要输入参数是model和batch, 无返回值, 建议在内部存储当前batch的值
            * `evaluate`方法, 无形参, return的是一个Dict[str, float]类型的评估结果
        period : int, default 1.
            执行eval_func函数的周期
        max_to_keep : int, 保存checkpoints的数量, 更早期的checkpoints会被删除
        save_metric : int, default None.
            保存模型的指标是哪个, 需要从trainer.metric_storage选择
        max_first : bool, default True.
            用于保存模型的指标是取最大还是最小作为最优模型
        save_last : bool, default True
            是否保存最近一次的epoch的模型, 如果是True, 每轮将更新模型到latest.pth中
        """
        self.prefix = prefix+"_"
        super(EvalTotalHook, self).__init__(period, max_to_keep, self.prefix + save_metric, max_first, save_last)
        assert hasattr(eval_metric, "update") and isinstance(getattr(eval_metric, "update"), Callable)
        assert hasattr(eval_metric, "evaluate") and isinstance(getattr(eval_metric, "evaluate"), Callable)
        self._eval_metric = eval_metric
        self.dataloader = dataloader

    @torch.no_grad()
    def _do_eval(self):
        mode_model = self.trainer.model_evaluate.training
        self.trainer.model_evaluate.eval()
        with tqdm(self.dataloader, desc="eval") as pbar:
            for batch in pbar:
                self._eval_metric.update(self.trainer.model_evaluate, batch)
        self.trainer.model_evaluate.train(mode_model)
        tot_res = self._eval_metric.evaluate()
        self._eval_metric.reset()
        if tot_res:
            rename_res = {self.prefix + k: np.mean(v) for k, v in tot_res.items()}
            self.log(self.trainer.epoch, **rename_res, smooth=False, window_size=1)

    def after_epoch(self):
        if self.every_n_epochs(self._period) or self.is_last_epoch():
            self._do_eval()
            self.save_model()


"""
该脚本主要用于测试计算资源对模型推理的速度
"""
from torchvision.models.resnet import resnet18, resnet50, resnet101, resnet152
from torchvision.models.mobilenet import mobilenet_v2
from torchvision.models.vgg import vgg16
from .progress_bar import ProgressBar
import torch
import time
import pandas as pd


BENCHMARK = {
    (16, 3, 224, 224): {
        "mobilenet_v2": {
            "RTX-3090": 6,
            "V100": 12,
            "A100": 8,
        },
        "resnet50": {
            "RTX-3090": 7,
            "V100": 9,
            "A100": 8
        },
        "vgg16": {
            "RTX-3090": 9,
            "V100": 7,
            "A100": 5
        },
    },
    (64, 3, 224, 224): {
        "mobilenet_v2": {
            "RTX-3090": 16,
            "V100": 25,
            "A100": 24,
        },
        "resnet50": {
            "RTX-3090": 19,
            "V100": 33,
            "A100": 24,
        },
        "vgg16": {
            "RTX-3090": 29,
            "V100": 28,
            "A100": 22,
        },
    }
}


def gpu_cnn_speed(model_type: str = None, input_size: tuple = (16, 3, 224, 224), gpu_id: int = None):
    """
    测试gpu在cnn上的性能

    Parameters
    ----------
    model_type : str, default resnet50. 模型类型
        * vgg16
        * resnet18
        * resnet50
        * resnet101
        * resnet152
        * mobilenet_v2
    input_size : tuple, default (16, 3, 224, 224). 模型输入尺寸
    gpu_id : int, default None. 显卡号，如果是None，且显卡可用，则默认为0
    """
    if model_type is None:
        model_type = "resnet50"
    assert model_type in ("vgg16", "resnet18", "resnet50", "resnet101", "resnet152", "mobilenet_v2")

    if gpu_id is None:
        gpu_id = 0
    device = f"cuda:{gpu_id}" if torch.cuda.is_available() else "cpu"
    model = eval(model_type)().to(device)
    iters = 300

    begin = time.time()
    pbar = ProgressBar(total=iters, desc=model_type)
    for i in range(iters):
        x = torch.randn(input_size).to(device)
        _ = model(x)
        pbar.update(1)
    print(f"全程耗时: {int(time.time() - begin + 0.5)}s")
    print("参考各显卡benchmark")
    for bis, v in BENCHMARK.items():
        print("*" * 50)
        print("benchmark input size:", bis)
        print(pd.DataFrame(v))



from __future__ import annotations

from typing import Optional, List

import numpy as np
import torch
import torchvision

__all__ = [
    "filter_box",
    "det_postprocess",
    "bboxes_iou",
    "xyxy2cxcywh",
    "box_candidates",
    "eval_map"
]


def filter_box(boxes: np.ndarray | torch.Tensor,
               min_scale: float = None,
               max_scale: float = None) -> np.ndarray | torch.Tensor:
    """
    目标框boxes根据尺寸做过滤
    Parameters
    ----------
    boxes : np.ndarray or torch.Tensor
        待过滤的目标框
    min_scale : float
    max_scale : float

    Returns
    -------
    boxes : np.ndarray or torch.Tensor
    """
    if min_scale is None:
        min_scale = 0.0
    if max_scale is None:
        max_scale = float('inf')
    w = boxes[:, 2] - boxes[:, 0]
    h = boxes[:, 3] - boxes[:, 1]
    keep = (w * h > min_scale * min_scale) & (w * h < max_scale * max_scale)
    return boxes[keep]


def det_postprocess(prediction: torch.Tensor,
                    num_classes: int,
                    conf_thre: float = 0.7,
                    nms_thre: float = 0.45,
                    class_agnostic: bool = False,
                    merge_score: bool = False,
                    ret_np: bool = False
                    ) -> List[Optional[torch.Tensor]]:
    """
    目标检测的后处理, 给定候选目标框, 经过置信度和mns双重筛选, 得到最终目标框
    Parameters
    ----------
    prediction : torch.Tensor
        检测模型给出的候选目标框, 采取了YOLO中boxes的形式: shape=(batch_size, num_boxes, 4 + 1 + num_classes)
        boxes的格式是x1,y1,x2,y2,confidence,class_score1, class_score2,...,class_scoreN
    num_classes : int
    conf_thre : float, default 0.7
        置信度阈值, 目标框的置信度分数=confidence*class_score, 过滤掉所有置信度分数小于conf_thre的boxes
    nms_thre : float, default 0.45
        mns的阈值
    class_agnostic: bool, default False
    merge_score: bool, default False. 如果是True则会把obj_score和cls_score合并,返回的output是
        [x1,y1, x2, y2, obj_score*cls_score, cls]
    ret_np: bool, default False. 是否返回np.ndarray类型结果

    Returns
    -------
    output : List[Optional[torch.Tensor]]
        做完置信度和mns过滤的boxes
    """
    box_corner = prediction.new(prediction.shape)
    box_corner[:, :, 0] = prediction[:, :, 0] - prediction[:, :, 2] / 2
    box_corner[:, :, 1] = prediction[:, :, 1] - prediction[:, :, 3] / 2
    box_corner[:, :, 2] = prediction[:, :, 0] + prediction[:, :, 2] / 2
    box_corner[:, :, 3] = prediction[:, :, 1] + prediction[:, :, 3] / 2
    prediction[:, :, :4] = box_corner[:, :, :4]

    output = [None for _ in range(len(prediction))]
    for i, image_pred in enumerate(prediction):

        # If none are remaining => process next image
        if not image_pred.size(0):
            continue
        # Get score and class with highest confidence
        class_conf, class_pred = torch.max(image_pred[:, 5: 5 + num_classes], 1, keepdim=True)

        conf_mask = (image_pred[:, 4] * class_conf.squeeze() >= conf_thre).squeeze()
        # Detections ordered as (x1, y1, x2, y2, obj_conf, class_conf, class_pred)
        detections = torch.cat((image_pred[:, :5], class_conf, class_pred.float()), 1)
        detections = detections[conf_mask]
        if not detections.size(0):
            continue

        if class_agnostic:
            nms_out_index = torchvision.ops.nms(
                detections[:, :4],
                detections[:, 4] * detections[:, 5],
                nms_thre,
            )
        else:
            nms_out_index = torchvision.ops.batched_nms(
                detections[:, :4],
                detections[:, 4] * detections[:, 5],
                detections[:, 6],
                nms_thre,
            )

        detections = detections[nms_out_index]
        if merge_score:
            detections = torch.cat(
                [
                    detections[:, :4],
                    torch.unsqueeze(detections[:, 4] * detections[:, 5], 1),
                    torch.unsqueeze(detections[:, 6], 1)
                ],
                dim=1
            )
        if output[i] is None:
            output[i] = detections
        else:
            output[i] = torch.cat((output[i], detections))
        if ret_np:
            output[i] = output[i].cpu().numpy()
    return output


def bboxes_iou_torch(
        bboxes_a: torch.Tensor,
        bboxes_b: torch.Tensor,
        xyxy: bool = True,
        mode: str = "iou",
        eps: float = 1e-6
):
    """
    Parameters
    ----------
    bboxes_a : torch.Tensor, shape=(batch_size, 4)
    bboxes_b : torch.Tensor, shape=(batch_size, 4)
    xyxy : bool, default True
        输入格式是[x1,y1,x2,y2]还是[cx,cy,w,h]
    mode : str, default `iou`. 支持在以下两种种类型中选择
        * `iou`
        * `iof`
    eps : float, default 1e-6. 为防止除以0引入的小数

    Returns
    -------
    iou : np.ndarray | torch.Tensor, shape=(batch_size, )
    """
    if bboxes_a.shape[1] != 4 or bboxes_b.shape[1] != 4:
        raise IndexError

    if xyxy:
        tl = torch.max(bboxes_a[:, None, :2], bboxes_b[:, :2])
        br = torch.min(bboxes_a[:, None, 2:], bboxes_b[:, 2:])
        area_a = torch.prod(bboxes_a[:, 2:] - bboxes_a[:, :2], 1)
        area_b = torch.prod(bboxes_b[:, 2:] - bboxes_b[:, :2], 1)
    else:
        tl = torch.max(
            (bboxes_a[:, None, :2] - bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] - bboxes_b[:, 2:] / 2),
        )
        br = torch.min(
            (bboxes_a[:, None, :2] + bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] + bboxes_b[:, 2:] / 2),
        )

        area_a = torch.prod(bboxes_a[:, 2:], 1)
        area_b = torch.prod(bboxes_b[:, 2:], 1)
    en = (tl < br).type(tl.type()).prod(dim=2)
    area_i = torch.prod(br - tl, 2) * en  # * ((tl < br).all())

    if mode == 'iou':
        return area_i / torch.clip(area_a[:, None] + area_b - area_i, eps)
    elif mode == 'iof':
        return area_i / torch.clip(torch.minimum(area_a[:, None], area_b), eps)
    else:
        raise ValueError("mode没有这种模式")


def bboxes_iou_numpy(
        bboxes_a: np.ndarray,
        bboxes_b: np.ndarray,
        xyxy: bool = True,
        mode: str = "iou",
        eps: float = 1e-6
):
    """
    Parameters
    ----------
    bboxes_a : np.ndarray | torch.Tensor, shape=(batch_size, 4)
    bboxes_b : np.ndarray | torch.Tensor, shape=(batch_size, 4)
    xyxy : bool, default True
        输入格式是[x1,y1,x2,y2]还是[cx,cy,w,h]
    mode : str, default `iou`. 支持在以下两种种类型中选择
        * `iou`
        * `iof`
    eps : float, default 1e-6. 为防止除以0引入的小数

    Returns
    -------
    iou : np.ndarray | torch.Tensor, shape=(batch_size, )
    """
    if bboxes_a.shape[1] != 4 or bboxes_b.shape[1] != 4:
        raise IndexError

    if xyxy:
        tl = np.maximum(bboxes_a[:, None, :2], bboxes_b[:, :2])
        br = np.minimum(bboxes_a[:, None, 2:], bboxes_b[:, 2:])
        area_a = np.prod(bboxes_a[:, 2:] - bboxes_a[:, :2], 1)
        area_b = np.prod(bboxes_b[:, 2:] - bboxes_b[:, :2], 1)
    else:
        tl = np.maximum(
            (bboxes_a[:, None, :2] - bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] - bboxes_b[:, 2:] / 2),
        )
        br = np.minimum(
            (bboxes_a[:, None, :2] + bboxes_a[:, None, 2:] / 2),
            (bboxes_b[:, :2] + bboxes_b[:, 2:] / 2),
        )

        area_a = np.prod(bboxes_a[:, 2:], 1)
        area_b = np.prod(bboxes_b[:, 2:], 1)
    en = (tl < br).astype(float).prod(axis=2)
    area_i = np.prod(br - tl, 2) * en  # * ((tl < br).all())

    if mode == 'iou':
        return area_i / np.maximum(area_a[:, None] + area_b - area_i, eps)
    elif mode == 'iof':
        return area_i / np.maximum(np.minimum(area_a[:, None], area_b), eps)
    else:
        raise ValueError("mode没有这种模式")


def bboxes_iou(
        bboxes_a: np.ndarray | torch.Tensor,
        bboxes_b: np.ndarray | torch.Tensor,
        xyxy: bool = True,
        mode: str = "iou",
        eps: float = 1e-6
):
    if isinstance(bboxes_a, np.ndarray) and isinstance(bboxes_b, np.ndarray):
        return bboxes_iou_numpy(bboxes_a, bboxes_b, xyxy, mode, eps)
    elif isinstance(bboxes_a, torch.Tensor) and isinstance(bboxes_b, torch.Tensor):
        return bboxes_iou_torch(bboxes_a, bboxes_b, xyxy, mode, eps)
    else:
        raise TypeError("`bboxes_a`和`bboxes_b`应为np.ndarray或torch.Tensor, 且相互类型一致")


def xyxy2cxcywh(bboxes: np.ndarray | torch.Tensor) -> np.ndarray | torch.Tensor:
    """
    输入boxes坐标, 格式从[[x1, y1, x2, y2]] -> [[cx, cy, w, h]], shape=[None, 4]

    Parameters
    ----------
    bboxes : np.ndarray or torch.Tensor
        目标框坐标数组, shape=(None, 4), 格式是[[x1, y1, x2, y2]]
    Returns
    -------
    bboxes : np.ndarray or torch.Tensor
        目标框坐标数组, shape=(None, 4), 格式是[[cx, cy, w, h]]
    """
    bboxes[:, 2] = bboxes[:, 2] - bboxes[:, 0]
    bboxes[:, 3] = bboxes[:, 3] - bboxes[:, 1]
    bboxes[:, 0] = bboxes[:, 0] + bboxes[:, 2] * 0.5
    bboxes[:, 1] = bboxes[:, 1] + bboxes[:, 3] * 0.5
    return bboxes


def box_candidates(box1: np.ndarray, box2: np.ndarray, wh_thr: int = 2, ar_thr: int = 20,
                   area_thr: float = 0.2):
    """
    对box2做过滤，对于一些极端情况做删除

    Parameters
    ----------
    box1 : np.ndarray. shape=(None, 4), 格式是[[x1, y1, x2, y2]].
    box2 : np.ndarray. shape=(None, 4), 格式是[[x1, y1, x2, y2]].
    wh_thr : int, default 2. 宽高的最小阈值, 小于该值会被过滤.
    ar_thr : int, default 20. 最大宽高比阈值, 大于该值会被过滤.
    area_thr: float, default 0.2. 与box1的面积比例，小于该值会被过滤

    Returns
    -------
    mask: np.ndarray. shape=(None, ), box2的mask.
    """
    # box1(4,n), box2(4,n)
    # Compute candidate boxes which include follwing 5 things:
    # box1 before augment, box2 after augment, wh_thr (pixels), aspect_ratio_thr, area_ratio
    w1, h1 = box1[:, 2] - box1[:, 0], box1[:, 3] - box1[:, 1]
    w2, h2 = box2[:, 2] - box2[:, 0], box2[:, 3] - box2[:, 1]
    ar = np.maximum(w2 / (h2 + 1e-16), h2 / (w2 + 1e-16))  # aspect ratio
    return (
            (w2 > wh_thr)
            & (h2 > wh_thr)
            & (w2 * h2 / (w1 * h1 + 1e-16) > area_thr)
            & (ar < ar_thr)
    )


def cal_match(pred_box, gt_box):
    count = 0
    iou = bboxes_iou(pred_box[:, 1:], gt_box[:, 1:])
    for i, pbox in enumerate(pred_box):
        for j, gbox in enumerate(gt_box.copy()):
            if iou[i][j] >= 0.9 and np.abs(pbox[1:] - gbox[1:]).max() < 5 and pbox[0] == gbox[0]:
                count += 1
                gt_box = np.concatenate([gt_box[:j], gt_box[j + 1:]])
                iou = np.concatenate([iou[:, :j], iou[:, j + 1:]], 1)
                break
    return count / len(pred_box)


def calc_ap(recall, precision, use_07_metric=True):
    # 是否使用 07 年的 11 点均值方式计算 ap
    if use_07_metric:
        ap = 0.
        for threshold in np.arange(0., 1.1, 0.1):
            if np.sum(recall >= threshold) == 0:
                p = 0
            else:
                p = np.max(precision[recall >= threshold])
            ap = ap + p / 11.
    else:
        # 增加哨兵，然后算出准确率包络
        mrec = np.concatenate(([0.], recall, [1.]))
        mpre = np.concatenate(([0.], precision, [0.]))
        for i in range(mpre.size - 1, 0, -1):
            mpre[i - 1] = np.maximum(mpre[i - 1], mpre[i])

        # 计算 pr 曲线中，召回率变化的下标
        idx = np.where(mrec[1:] != mrec[:-1])[0]

        # 计算 pr 曲线与坐标轴所围区域的面积
        ap = np.sum((mrec[idx + 1] - mrec[idx]) * mpre[idx + 1])

    return ap


def clac_voc_ap(annotation, prediction, ovthresh=0.5):
    # 统计标注框个数，这里需要将困难样本数量去掉
    positive_num = 0
    for anno in annotation:
        positive_num += len(anno.get("bboxes", []))
        anno['b_det'] = [False] * len(anno.get("bboxes", []))

    # 将检测结果格式进行转换，主要是为了方便排序
    # for item in prediction:
    image_ids, confidences, bboxes = [], [], []
    for img_id, item in enumerate(prediction):
        bbox = item.get("bboxes", [])
        score = item.get("confidence", [])
        for i in range(len(score)):
            image_ids.append(img_id)
            confidences.append(score[i])
            bboxes.append(bbox[i])
    image_ids, confidences, bboxes = np.array(image_ids), np.array(confidences), np.array(bboxes)

    # 按照置信度排序
    sorted_ind = np.argsort(-confidences)
    bboxes = bboxes[sorted_ind]
    image_ids = image_ids[sorted_ind]

    # 计算 TP 和 FP，以计算出 AP 值
    detect_num = len(image_ids)
    tp = np.zeros(detect_num)
    fp = np.zeros(detect_num)
    for d in range(detect_num):
        gt_bboxes = annotation[image_ids[d]].get("bboxes", [])

        # 如果没有 ground truth，那么所有的检测都是错误的
        if len(gt_bboxes) > 0:
            b_dets = annotation[image_ids[d]]["b_det"]
            p_bboxes = bboxes[d, :]

            overlaps = bboxes_iou(p_bboxes[np.newaxis], gt_bboxes, mode="iou")[0]
            idxmax = np.argmax(overlaps)
            ovmax = overlaps[idxmax]

            if ovmax > ovthresh:
                # gt 只允许检测出一次
                if not b_dets[idxmax]:
                    tp[d] = 1.
                    b_dets[idxmax] = True
                else:
                    fp[d] = 1.
            else:
                fp[d] = 1.
        else:
            fp[d] = 1.

    # 计算召回率和准确率
    tp = np.cumsum(tp)
    fp = np.cumsum(fp)
    recall = tp / float(positive_num)
    precision = tp / np.maximum(tp + fp, 1.)
    ap = calc_ap(recall, precision)

    return recall, precision, ap


def eval_map(annotation_result, detector_result, ovthresh=0.5):
    aps = {}
    for cls in annotation_result.keys():
        recall, precision, ap = clac_voc_ap(annotation_result[cls], detector_result[cls], ovthresh=ovthresh)
        aps[cls] = ap
    return aps



import logging
import os
import random
import sys
from collections import defaultdict
from typing import Any, Dict

import datetime
import numpy as np
import torch
from tabulate import tabulate

logger = logging.getLogger(__name__)

__all__ = [
    "set_random_seed",
    "collect_env",
    "symlink",
    "create_small_table",
    "get_workspace"
]


def collect_env() -> str:
    """Collect the information of the running environments.

    The following information are contained.

        - sys.platform: The variable of ``sys.platform``.
        - Python: Python version.
        - Numpy: Numpy version.
        - CUDA available: Bool, indicating if CUDA is available.
        - GPU devices: Device type of each GPU.
        - PyTorch: PyTorch version.
        - PyTorch compiling details: The output of ``torch.__config__.show()``.
        - TorchVision (optional): TorchVision version.
        - OpenCV (optional): OpenCV version.

    Returns:
        str: A string describing the running environment.
    """
    env_info = []
    env_info.append(("sys.platform", sys.platform))
    env_info.append(("Python", sys.version.replace("\n", "")))
    env_info.append(("Numpy", np.__version__))

    cuda_available = torch.cuda.is_available()
    env_info.append(("CUDA available", cuda_available))

    if cuda_available:
        devices = defaultdict(list)
        for k in range(torch.cuda.device_count()):
            devices[torch.cuda.get_device_name(k)].append(str(k))
        for name, device_ids in devices.items():
            env_info.append(("GPU " + ",".join(device_ids), name))

    env_info.append(("PyTorch", torch.__version__))

    try:
        import torchvision

        env_info.append(("TorchVision", torchvision.__version__))
    except ModuleNotFoundError:
        pass

    try:
        import cv2

        env_info.append(("OpenCV", cv2.__version__))
    except ModuleNotFoundError:
        pass

    torch_config = torch.__config__.show()
    env_str = tabulate(env_info) + "\n" + torch_config
    return env_str


def set_random_seed(seed: int, rank: int = 0) -> None:
    """Set random seed.

    Args:
        seed (int): Nonnegative integer.
        rank (int): Process rank in the distributed training. Defaults to 0.
    """
    assert seed >= 0, f"Got invalid seed value {seed}."
    seed += rank
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    np.random.seed(seed)  # Numpy module.
    random.seed(seed)  # Python random module.
    torch.backends.cudnn.benchmark = False  # 保证每次卷积的算子都是固定的，而非使用最高效的方法
    torch.backends.cudnn.deterministic = True

    os.environ["PYTHONHASHSEED"] = str(seed)


def symlink(src: str, dst: str, overwrite: bool = True, **kwargs) -> None:
    """Create a symlink, dst -> src.

    Args:
        src (str): Path to source.
        dst (str): Path to target.
        overwrite (bool): If True, remove existed target. Defaults to True.
    """
    if os.path.lexists(dst) and overwrite:
        os.remove(dst)
    os.symlink(src, dst, **kwargs)


def create_small_table(small_dict: Dict[str, Any]) -> str:
    """Create a small table using the keys of ``small_dict`` as headers.
    This is only suitable for small dictionaries.

    Args:
        small_dict (dict): A result dictionary of only a few items.

    Returns:
        str: The table as a string.
    """
    keys, values = tuple(zip(*small_dict.items()))
    table = tabulate(
        [values],
        headers=keys,
        tablefmt="pipe",
        floatfmt=".3f",
        stralign="center",
        numalign="center",
    )
    return table


def get_workspace(work_dir, create_new_dir):
    if create_new_dir not in (None, "time_s", "time_m", "time_h", "time_d", "count"):
        logger.warning("create_new_dir参数输入错误, 使用`time_s`为其赋值")
        create_new_dir = "time_s"
    if os.path.exists(work_dir):
        if create_new_dir == "time_s":
            now = datetime.datetime.now()
            now_format = now.strftime("%Y-%m-%d %H_%M_%S")
            work_dir = f"{work_dir}_{now_format}"
        elif create_new_dir == "time_m":
            now = datetime.datetime.now()
            now_format = now.strftime("%Y-%m-%d %H_%M")
            work_dir = f"{work_dir}_{now_format}"
        elif create_new_dir == "time_h":
            now = datetime.datetime.now()
            now_format = now.strftime("%Y-%m-%d %H")
            work_dir = f"{work_dir}_{now_format}"
        elif create_new_dir == "time_d":
            now = datetime.datetime.now()
            now_format = now.strftime("%Y-%m-%d")
            work_dir = f"{work_dir}_{now_format}"
        elif create_new_dir == "count":
            for i in range(10000):
                if not os.path.exists(f"{work_dir}_{i}"):
                    break
            work_dir = f"{work_dir}_{i}"
    return work_dir



from accelerate import Accelerator
from accelerate.optimizer import AcceleratedOptimizer
from accelerate.utils import ProjectConfiguration, set_seed
import warnings
from torch import nn
import torch
import torch.optim as optim
from torch.utils.data import DataLoader
from typing import Optional, Union, List, Dict, Any
from .trainer import Trainer, MetricStorage, osp
from .hooks import CheckpointerHook, HookBase, LoggerHook
from .utils.misc import get_workspace
from .lr_scheduler import LRWarmupScheduler
from ._get_logger import logger


class AccelerateTrainer(Trainer):
    """
    accelerate版的Trainer,底层以来accelerate库
    其他参数都和Trainer一致, 除了以下参数
    Parameters
    ---------
    lr_scheduler: Union[str, optim.lr_scheduler._LRScheduler]. 会有以下差别:
        增加str类型，这时会直接调用diffusers内部的lr_scheduler, 这里建议使用str类型
        支持["linear", "cosine", "cosine_with_restarts", "polynomial", "constant", "constant_with_warmup"]
    mixed_precision: str, default None
         * "no", 不使用amp技术
         * "fp16"
         * "bf16", 30系及以后的卡型才能使用
    """

    def __init__(
            self,
            model: nn.Module,
            optimizer: optim.Optimizer,
            lr_scheduler: Union[str, optim.lr_scheduler._LRScheduler],
            data_loader: DataLoader,
            max_epochs: int,
            work_dir: str = "work_dir",
            clip_grad_norm: float = 0.0,
            mixed_precision: str = None,
            warmup_method: Optional[str] = None,
            warmup_iters: int = 0,
            warmup_factor: float = 0.001,
            hooks: Optional[List[HookBase]] = None,
            use_ema: bool = False,
            ema_decay: float = 0.9999,
            create_new_dir: Optional[str] = "time_s"
    ):
        self.work_dir = get_workspace(work_dir, create_new_dir)
        accelerator_project_config = ProjectConfiguration(project_dir=self.work_dir, logging_dir=self.work_dir)
        self.accelerator = Accelerator(
            mixed_precision=mixed_precision,
            project_config=accelerator_project_config,
        )

        self.weight_dtype = torch.float32
        if self.accelerator.mixed_precision == "fp16":
            self.weight_dtype = torch.float16
        elif self.accelerator.mixed_precision == "bf16":
            self.weight_dtype = torch.bfloat16

        self.model = model
        if use_ema:
            warnings.warn("accelerate版暂不支持ema, 敬请期待")
        self.model_ema = None
        self.optimizer = optimizer
        # convert epoch-based scheduler to iteration-based scheduler
        if isinstance(lr_scheduler, str):
            from diffusers.optimization import get_scheduler
            if warmup_method is not None:
                warnings.warn("当`lr_scheduler`输入str类型时, `warmup_method`参数失效")
            max_train_steps = max_epochs * len(data_loader)
            self.lr_scheduler = get_scheduler(
                lr_scheduler,
                optimizer=optimizer,
                num_warmup_steps=warmup_iters * self.accelerator.num_processes,
                num_training_steps=max_train_steps * self.accelerator.num_processes,
            )
        else:
            self.lr_scheduler = LRWarmupScheduler(
                lr_scheduler, len(data_loader), warmup_method, warmup_iters, warmup_factor
            )
        self.data_loader = data_loader
        self.metric_storage = MetricStorage()

        # counters
        self.inner_iter: int = -1  # [0, epoch_len - 1]
        self.epoch: int = -1  # [0, max_epochs - 1]
        self.start_epoch = 0  # [0, max_epochs - 1]
        self.max_epochs = max_epochs

        self._hooks: List[HookBase] = []
        self._clip_grad_norm = clip_grad_norm

        if self.accelerator.is_main_process:
            if hooks is None:
                hooks = self._build_default_hooks()
            self.register_hooks(hooks)

        self._prepared_status = False

    def prepare_model(self):
        """如果有多个模型的话可以在这里重写方法"""
        self.model, self.optimizer, self.data_loader, self.lr_scheduler = self.accelerator.prepare(
            self.model, self.optimizer, self.data_loader, self.lr_scheduler
        )
        self._prepared_status = True

    def _prepare_for_training(self,
                              console_log_level: int = 2,
                              file_log_level: int = 2) -> None:
        """
        训练前的配置工作
        Parameters
        ----------
        console_log_level : int, default 2
             输出到屏幕的log等级, 可选范围是0-5, 它们对应的关系分别为：
             * 5: FATAL
             * 4: ERROR
             * 3: WARNING
             * 2: INFO
             * 1: DEBUG
             * 0: NOTSET
        file_log_level : int, default 2
             输出到文件里的log等级, 其他方面同console_log_level参数
        """
        self.prepare_model()  # 此处和原版trainer不同!!!
        super()._prepare_for_training(console_log_level, file_log_level)

    def train_one_iter(self, batch) -> dict:
        """
        包含了accelerate版训练的一个iter的全部操作

        .. Note::
            标准的学习率调节器是基于epoch的, 但torch_frame框架是基于iter的, 所以它在每次iter之后都会调用
        """
        #####################
        # 1. 计算loss #
        #####################
        loss_info = self.model(batch)
        if isinstance(loss_info, torch.Tensor):
            losses = loss_info
            loss_info = {"total_loss": loss_info}
        elif isinstance(loss_info, tuple):
            assert len(loss_info) == 2, "loss_info需要是一个二元组，第一个是需要反向传播的，第二个是其他参考指标"
            backward_params, metric_params = loss_info
            assert isinstance(metric_params, dict), "loss_info的第二个值需要是dict类型"
            if isinstance(backward_params, torch.Tensor):
                losses = backward_params
                metric_params["total_loss"] = backward_params
                loss_info = metric_params
            elif isinstance(backward_params, dict):
                losses = sum(backward_params.values())
                backward_params["total_loss"] = losses
                backward_params.update(metric_params)
                loss_info = backward_params
        else:
            assert "total_loss" not in loss_info, "当model返回是一个dict的时候不可以传出包含"
            losses = sum(loss_info.values())
            loss_info["total_loss"] = losses

        ##########################
        # 2. 计算梯度 #
        ##########################
        self.optimizer.zero_grad()
        self.accelerator.backward(losses)
        if self._clip_grad_norm > 0 and self.accelerator.sync_gradients:
            self.accelerator.clip_grad_norm_(self.model.parameters(), self._clip_grad_norm)

        ##############################
        # 3. 更新模型参数 #
        ##############################
        self.optimizer.step()
        if self.model_ema:
            self.model_ema.update(self.model)

        ###########################
        # 4. 调整学习率 #
        ###########################
        self.lr_scheduler.step()

        show_info = {k: v.detach().cpu().item() if isinstance(v, torch.Tensor) else v for k, v in loss_info.items()}
        show_info = dict(sorted(show_info.items(), key=lambda x: x[0] != "total_loss"))  # 保证total_loss在最后一位
        self.pbar.set_postfix(show_info)
        return show_info

    @property
    def model_or_module(self) -> nn.Module:
        if self.model.__class__.__name__ == "DistributedDataParallel":
            return self.accelerator.unwrap_model(self.model)
        return self.model



from .config_parser import ConfigArgumentParser, save_args
from .misc import *
from .logger import setup_logger
from .history_buffer import HistoryBuffer
from .progress_bar import ProgressBar
from .ema import EMA



import os
import logging
import argparse
import yaml
from copy import deepcopy
from argparse import Namespace
from typing import Optional, List

logger = logging.getLogger(__name__)


class ConfigArgumentParser(argparse.ArgumentParser):
    """Argument parser that supports loading a YAML configuration file.

    A small issue: config file values are processed using ArgumentParser.set_defaults(..)
    which means "required" and "choices" are not handled as expected. For example, if you
    specify a required value in a config file, you still have to specify it again on the
    command line. The ``ConfigArgParse`` library (http://pypi.python.org/pypi/ConfigArgParse)
    can be used as a substitute.
    """

    def __init__(self, *args, **kwargs):
        self.config_parser = argparse.ArgumentParser(add_help=False)
        self.config_parser.add_argument("-c", "--config", default=None, metavar="FILE",
                                        help="where to load YAML configuration")
        self.option_names = []
        super().__init__(*args,
                         # Inherit options from config_parser
                         parents=[self.config_parser],
                         # Don't mess with format of description
                         formatter_class=argparse.RawDescriptionHelpFormatter,
                         **kwargs)

    def add_argument(self, *args, **kwargs):
        arg = super().add_argument(*args, **kwargs)
        self.option_names.append(arg.dest)
        return arg

    def parse_args(self, args=None):
        res, remaining_argv = self.config_parser.parse_known_args(args)

        if res.config is not None:
            with open(res.config, "r") as f:
                config_vars = yaml.safe_load(f)
            for key in config_vars:
                if key not in self.option_names:
                    self.error(f"unexpected configuration entry: {key}")
            self.set_defaults(**config_vars)

        return super().parse_args(remaining_argv)


def save_args(args: Namespace, filepath: str, excluded_fields: Optional[List[str]] = None) -> None:
    """Save args with some excluded fields to a ``.yaml`` file.

    Args:
        args (Namespace): The parsed arguments to be saved.
        filepath (str): A filepath ends with ".yaml".
        excluded_fields (list[str]): The names of some fields that are not saved.
            Defaults to ["config"].
    """
    assert isinstance(args, Namespace)
    assert filepath.endswith(".yaml")
    os.makedirs(os.path.dirname(os.path.abspath(filepath)), exist_ok=True)
    save_dict = deepcopy(args.__dict__)
    for field in excluded_fields or ["config"]:
        save_dict.pop(field)
    with open(filepath, "w") as f:
        yaml.dump(save_dict, f)
    logger.info(f"Args is saved to {filepath}")



from .trainer import Trainer, nn, optim, Optional, List, HookBase, setup_logger, logger
import os
from torch.utils.data import DataLoader, Dataset, DistributedSampler, RandomSampler
from .utils.dist_utils import *
import warnings


class DDPTrainer(Trainer):
    """
    Trainer的DDP版本。需要注意的是，在调用该类前依然要执行init_process_group，为了避免重复打印或保存数据，应该在外部选择性
    的根据gpu_id调用hooker, 在创建DDPTrainer对象时会自动根据环境判断是否使用ddp，无需手动设置


    Parameters
    ---------
    model : torch.nn.Module, 训练模型, 训练时的输出只能是以下三种:
        * torch.Tensor, 对于这种输出是模型backward用的loss, 在torch-frame框架中被称为total_loss
        * dict, 里面是模型的各路分支的loss，需要是标量, Trainer会自动将其求和得到total_loss, 再做backward
        * Tuple[Union[dict, torch.Tensor], dict]. 前两种的混合体, 元组第一个输出是前面两种的任意一种;
          第二个输出是非需要backward类的, Trainer不会把这个dict汇总到total_loss上
    optimizer : torch.optim.Optimizer, 优化器
    lr_scheduler : optim.lr_scheduler._LRScheduler, 学习率调节器
    dataset : torch.utils.data.Dataset, 训练集数据生成器, 不需要创建dataloader, 由DDPTrainer内部创建
    dataset_params : dict, dataset的参数, key是诸如batch_size的参数
    max_epochs : int, 训练的总轮数
    work_dir : str, 保存模型和日志的根目录地址
    clip_grad_norm : float, default 0.0
        梯度裁剪的设置, 如果置为小于等于0, 则不作梯度裁剪
    enable_amp : bool, 使用混合精度
    warmup_method : str, default None
        warmup的类型, 包含以下四种取值
        * constant
        * linear
        * exp
        * None : 不使用warmup
    warmup_iters : int, default 1000, warmup最后的iter数
    warmup_factor : float, default 0.001
        warmup初始学习率 = warmup_factor * initial_lr
    hooks : List[HookBase], default None.
        hooks, 保存模型、输出评估指标、loss等用
    use_ema : bool, default False. 是否使用EMA技术
    ema_decay: float = 0.9999. EMA模型衰减系数
    create_new_dir : Optional[str], default time
        存在同名目录时以何种策略创建目录
        * None, 直接使用同名目录
        * `time_s`, 如果已经存在同名目录, 则以时间(精确到秒)为后缀创建新目录
        * `time_m`, 如果已经存在同名目录, 则以时间(精确到分)为后缀创建新目录
        * `time_h`, 如果已经存在同名目录, 则以时间(精确到小时)为后缀创建新目录
        * `time_d`, 如果已经存在同名目录, 则以时间(精确到日)为后缀创建新目录
        * `count`, 如果已经存在同名目录, 则以序号为后缀创建新目录
    """

    def __init__(self,
                 model: nn.Module,
                 optimizer: optim.Optimizer,
                 lr_scheduler: optim.lr_scheduler._LRScheduler,
                 dataset: Dataset,
                 dataset_params: dict,
                 max_epochs: int,
                 work_dir: str = "work_dir",
                 clip_grad_norm: float = 0.0,
                 enable_amp=False,
                 warmup_method: Optional[str] = None,
                 warmup_iters: int = 1000,
                 warmup_factor: float = 0.001,
                 hooks: Optional[List[HookBase]] = None,
                 use_ema: bool = False,
                 ema_decay: float = 0.9999,
                 create_new_dir: Optional[str] = "time_s"
                 ):
        warnings.warn("为了更快训练，建议使用AccelerateTrainer来替代这个trainer", DeprecationWarning)
        self.use_dist = is_dist_avail_and_initialized()
        if self.use_dist:
            num_tasks = get_world_size()
            global_rank = get_rank()
            sampler_trainer = DistributedSampler(dataset, num_replicas=num_tasks, rank=global_rank)
        else:
            sampler_trainer = RandomSampler(dataset)
        data_loader = DataLoader(dataset, sampler=sampler_trainer, **dataset_params)
        super(DDPTrainer, self).__init__(model, optimizer, lr_scheduler, data_loader, max_epochs, work_dir,
                                         clip_grad_norm, enable_amp,  warmup_method, warmup_iters, warmup_factor,
                                         hooks, use_ema, ema_decay, create_new_dir)

    def _train_one_epoch(self) -> None:
        """执行模型一个epoch的全部操作"""
        if self.use_dist:
            # dist.barrier()  这里可能会造成阻塞
            self.data_loader.sampler.set_epoch(self.epoch)
        super(DDPTrainer, self)._train_one_epoch()

    def _prepare_for_training(self,
                              console_log_level: int = 2,
                              file_log_level: int = 2) -> None:
        """
        训练前的配置工作
        Parameters
        ----------
        console_log_level : int, default 2
             输出到屏幕的log等级, 可选范围是0-5, 它们对应的关系分别为：
             * 5: FATAL
             * 4: ERROR
             * 3: WARNING
             * 2: INFO
             * 1: DEBUG
             * 0: NOTSET
        file_log_level : int, default 2
             输出到文件里的log等级, 其他方面同console_log_level参数
        """
        assert console_log_level in (0, 1, 2, 3, 4, 5), f"console_log_level必须在0~5之间而不是{console_log_level}"
        assert file_log_level in (0, 1, 2, 3, 4, 5), f"file_log_level必须在0~5之间而不是{file_log_level}"
        console_log_level *= 10
        file_log_level *= 10
        setup_logger("torch_frame", output=self.log_file,
                     console_log_level=console_log_level, file_log_level=file_log_level)

        if self.start_epoch == 0 and get_rank() == 0:
            os.makedirs(self.ckpt_dir, exist_ok=True)
            logger.info(f"Registered default hooks: {self.registered_hook_names}")
            split_line = "-" * 50
            logger.info(
                f"\n{split_line}\n"
                f"Work directory: {self.work_dir}\n"
                f"{split_line}"
            )

        


import numpy as np
from collections import deque


class HistoryBuffer:
    """
    该类保存了一些数, 并且可以得到平滑均值

    Example::

        >>> his_buf = HistoryBuffer()
        >>> his_buf.update(0.1)
        >>> his_buf.update(0.2)
        >>> his_buf.avg
        0.15
    """

    def __init__(self, window_size: int = 20) -> None:
        """
        Parameters
        ----------
        window_size : int, default 20. 滑窗大小
        """
        self._history = deque(maxlen=window_size)
        self._count: int = 0
        self._sum: float = 0.0

    def update(self, value: float) -> None:
        """
        在列表新增变量, 如果新增后超出窗口大小, 舍去最早的那个值
        """
        self._history.append(value)
        self._count += 1
        self._sum += value

    @property
    def latest(self) -> float:
        return self._history[-1]

    @property
    def avg(self) -> float:
        return np.mean(self._history)

    @property
    def global_avg(self) -> float:
        return self._sum / self._count

    @property
    def global_sum(self) -> float:
        return self._sum

    def __le__(self, other):
        return self.avg <= other

    def __lt__(self, other):
        return self.avg < other

    def __ge__(self, other):
        return self.avg >= other

    def __gt__(self, other):
        return self.avg > other

    def __str__(self):
        return str(round(self.avg, 4))



from ._get_logger import logger
from .hooks import *
from .lr_scheduler import LRWarmupScheduler
from .utils import HistoryBuffer, misc, metric
from .trainer import Trainer, MetricStorage
from .ddp_trainer import DDPTrainer
from .accelerate_trainer import AccelerateTrainer


#  关闭opencv的多线程防止pytorch死锁
import cv2
cv2.setNumThreads(0)
cv2.ocl.setUseOpenCL(False)



from torch.utils.data.dataloader import DataLoader


__all__ = [
    "InfiniteDataLoader",
]


# 源自https://github.com/AlibabaResearch/efficientteacher/blob/main/utils/datasets_ssod.py
class InfiniteDataLoader(DataLoader):
    """ Dataloader that reuses workers

    Uses same syntax as vanilla DataLoader
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        object.__setattr__(self, 'batch_sampler', _RepeatSampler(self.batch_sampler))
        self.iterator = super().__iter__()

    def __len__(self):
        return len(self.batch_sampler.sampler)

    def __iter__(self):
        for i in range(len(self)):
            yield next(self.iterator)


class _RepeatSampler(object):
    """ Sampler that repeats forever

    Args:
        sampler (Sampler)
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            yield from iter(self.sampler)


import cv2
import random
import numpy as np


__all__ = [
    "mixup"
]


def mixup(img1: np.ndarray, img2: np.ndarray,
                mixup_scale: float = 0.1):
    def translate(image, h, w, alpha):
        y_offset = random.randint(0, dst_img.shape[0] - h)
        x_offset = random.randint(0, dst_img.shape[1] - w)
        dst_img[y_offset: h+y_offset, x_offset: w+x_offset] = image * alpha

    h1, w1 = img1.shape[:2]
    h2, w2 = img2.shape[:2]
    dst_img = np.zeros((max(h1, h2), max(w1, w2), 3), dtype=np.float)
    alpha = random.uniform(0.5 - mixup_scale, 0.5 + mixup_scale)
    img1 = img1.astype(np.float32)
    translate(img1, h1, w1, alpha)
    translate(img1, h2, w2, 1 - alpha)
    return dst_img.astype(np.uint8)


import math
from copy import deepcopy
import torch
import torch.nn as nn


class EMA:
    """
    Model Exponential Moving Average from https://github.com/rwightman/pytorch-image-models
    Keep a moving average of everything in the model state_dict (parameters and buffers).
    This is intended to allow functionality like
    https://www.tensorflow.org/api_docs/python/tf/train/ExponentialMovingAverage
    A smoothed version of the weights is necessary for some training schemes to perform well.
    This class is sensitive where it is initialized in the sequence of model init,
    GPU assignment and distributed training wrappers.
    """

    def __init__(self, model, decay=0.9999, updates=0):
        """
        Args:
            model (nn.Module): model to apply EMA.
            decay (float): ema decay reate.
            updates (int): counter of EMA updates.
        """
        # Create EMA(FP32)
        self.model = deepcopy(model).eval()
        self.updates = updates
        # decay exponential ramp (to help early epochs)
        self.decay = lambda x: decay * (1 - math.exp(-x / 2000))
        for p in self.model.parameters():
            p.requires_grad_(False)

    def update(self, model):
        # Update EMA parameters
        with torch.no_grad():
            self.updates += 1
            d = self.decay(self.updates)
            msd = model.state_dict()
            for k, v in self.model.state_dict().items():
                if v.dtype.is_floating_point:
                    v *= d
                    v += (1.0 - d) * msd[k].detach()



import datetime
import logging
import time
from typing import Dict, Optional

import torch
from torch.utils.tensorboard import SummaryWriter

from .hookbase import HookBase

logger = logging.getLogger(__name__)


class LoggerHook(HookBase):
    """写入评估指标到控制台和tensorboard"""

    def __init__(self, period: int = 50, tb_log_dir: Optional[str] = None, modes: Optional[list] = None, **kwargs) -> None:
        """
        Parameters
        ----------
        period : int, default 50. 写入的周期
        tb_log_dir : str, default None. 如果没有特别指定的话，日志默认写到trainer设置的目录
        modes : list, default None
            通过指标的关键字来判定输出哪类数据
        kwargs : torch.utils.tensorboard.SummaryWriter的其他参数
        """
        self._period = period
        self.kwargs = kwargs
        # metric name -> the latest iteration written to tensorboard file
        self._last_write: Dict[str, int] = {}

        if modes is None:
            modes = ["train", "eval"]
        if "train" not in modes:
            modes.insert(0, "train")
        self.modes = {m + "_" for m in modes}
        self.tb_log_dir = tb_log_dir

    def before_train(self) -> None:
        self._train_start_time = time.perf_counter()
        if self.tb_log_dir is None:
            self.tb_log_dir = self.trainer.work_dir
        self._tb_writer = SummaryWriter(self.tb_log_dir, **self.kwargs)

    def after_train(self) -> None:
        self._tb_writer.close()
        total_train_time = time.perf_counter() - self._train_start_time
        total_hook_time = total_train_time - self.metric_storage["iter_time"].global_sum
        logger.info(
            "Total train time: {} ({} on hooks)".format(
                str(datetime.timedelta(seconds=int(total_train_time))),
                str(datetime.timedelta(seconds=int(total_hook_time))),
            )
        )

    def after_epoch(self) -> None:
        self._write_console()
        self._write_tensorboard()

    def _write_console(self) -> None:
        # These fields ("data_time", "iter_time", "lr", "loss") may does not
        # exist when user overwrites `self.trainer.train_one_iter()`
        data_time = self.metric_storage["data_time"].avg if "data_time" in self.metric_storage else None
        iter_time = self.metric_storage["iter_time"].avg if "iter_time" in self.metric_storage else None
        lr = self.metric_storage["lr"].latest if "lr" in self.metric_storage else None

        if iter_time is not None:
            eta_seconds = iter_time * (self.trainer.max_iters - self.trainer.cur_iter - 1)
            eta_string = str(datetime.timedelta(seconds=int(eta_seconds)))
        else:
            eta_string = None

        if torch.cuda.is_available():
            max_mem_mb = torch.cuda.max_memory_allocated() / 1024.0 / 1024.0
        else:
            max_mem_mb = None

        exclude = ("data_time", "iter_time", "lr")
        keys_dict = {mode: set() for mode in self.modes}
        for key in self.metric_storage:
            if key in exclude:
                continue
            for mode in self.modes:
                if key.startswith(mode):
                    keys_dict[mode].add(key)
                    break
            else:
                keys_dict["train_"].add(key)

        process_string = f"Epoch: [{self.trainer.epoch}][{self.trainer.inner_iter}/{self.trainer.epoch_len - 1}]"

        space = " " * 2
        logger.info("----------")
        logger.info(
            "{process}{eta}{iter_time}{data_time}{lr}{memory}".format(
                process=process_string,
                eta=space + f"ETA: {eta_string}" if eta_string is not None else "",
                iter_time=space + f"iter_time: {iter_time:.4f}" if iter_time is not None else "",
                data_time=space + f"data_time: {data_time:.4f}  " if data_time is not None else "",
                lr=space + f"lr: {lr:.5g}" if lr is not None else "",
                memory=space + f"max_mem: {max_mem_mb:.0f}M" if max_mem_mb is not None else "",
            )
        )

        for keys in keys_dict.values():
            key_list = sorted(list(keys), key=lambda x: "total" not in x)
            info = " ".join([f"{k}: {self.metric_storage[k]}" for k in key_list])
            if info == "":
                continue
            logger.info(info)

    def _write_tensorboard(self) -> None:
        for key, (iter, value) in self.metric_storage.values_maybe_smooth.items():
            if key not in self._last_write or iter > self._last_write[key]:
                for mode in self.modes:
                    if key.startswith(mode):
                        key = f"{mode.strip('_')}/{key}"
                        break
                else:
                    key = f"train/{key}"
                self._tb_writer.add_scalar(key, value, iter)
                self._last_write[key] = iter

    def after_iter(self) -> None:
        if self.every_n_inner_iters(self._period):
            # self._write_console()
            self._write_tensorboard()



from random import randint, random, uniform
import numpy as np
import cv2

__all__ = [
    "RandomBrightness",
    "RandomGammaCorrection",
    "RandomHueSaturation",
    "RandomContrast",
]


class RandomBrightness:
    def __init__(self, low=-10, high=10, p=0.2):
        if low < -80 or high > 80:
            raise ValueError("亮度low不可以小于-80, high不可以大于80")
        self.low = low
        self.high = high
        self.p = p

    def __call__(self, image):
        if random() < self.p:
            return image
        image = image.astype(np.float32)
        brightness = randint(self.low, self.high)
        image[:, :, :] += brightness
        image = np.clip(image, 0, 255).astype(np.uint8)
        return image


class RandomGammaCorrection:
    def __init__(self, lower=0.666, upper=1.5, p=0.5):
        self.lower = lower
        self.upper = upper
        self.p = p

    def __call__(self, image):
        if random() < self.p:
            return image
        cur_gamma = uniform(self.lower, self.upper)
        max_channels = (255, 255, 255)
        image = np.power(image / max_channels, cur_gamma)
        image = (image * 255).astype(np.uint8)
        return image


class RandomHueSaturation:
    def __init__(self, hue_lower=-12, hue_upper=12, sat_lower=0.6, sat_upper=1.5, p=0.5):
        self.hue_lower = hue_lower
        self.hue_upper = hue_upper
        self.sat_lower = sat_lower
        self.sat_upper = sat_upper
        self.p = p

    def __call__(self, image):
        if random() < self.p:
            return image
        random_h = uniform(self.hue_lower, self.hue_upper)
        random_s = uniform(self.sat_lower, self.sat_upper)
        image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV).astype(np.float32)
        h = image[:, :, 0]
        h += random_h
        h[h > 180.0] -= 180.0
        h[h < 0.0] += 180.0

        s = image[:, :, 1]
        s *= random_s
        # 一定要先转uint8
        image = cv2.cvtColor(image.astype(np.uint8), cv2.COLOR_HSV2BGR)
        return image


class RandomContrast:
    def __init__(self, low=0.9, high=1.1, p=0.2):
        if low < 0.5 or high > 2:
            raise ValueError("对比度low不可以小于0.5, high不可以大于2.0")
        if low >= high:
            raise ValueError("对比度low不可以大于或等于high")
        self.low = low
        self.high = high
        self.p = p

    def __call__(self, image):
        if random() < self.p:
            return image
        image = image.astype(np.float32)
        contrast = uniform(self.low, self.high)
        image[:, :, :] *= contrast
        image = np.clip(image, 0, 255).astype(np.uint8)
        return image


# from ..trainer import Trainer, MetricStorage
# import numpy as np


class HookBase:
    """
    hooks的基类

    Hook类在Trainer类中被初始化。每个Hook可以在六个阶段执行, 对应Trainer的六个方法分别为：

    * before_train, 训练前
    * after_train, 训练后
    * before_epoch, 一轮epoch前
    * after_epoch, 一轮epoch后
    * before_iter, 一个iter前
    * after_iter, 一个iter后
    目前在Hook中, 不能得到通过self.trainer得到类似model、optimizer等信息

    Examples:

    >>> hook.before_train()
    >>> for epoch in range(start_epoch, max_epochs):
    >>>     hook.before_epoch()
    >>>     for iter in range(epoch_len):
    >>>         hook.before_iter()
    >>>         train_one_iter()
    >>>         hook.after_iter()
    >>>     hook.after_epoch()
    >>> hook.after_train()
    """

    # A weak reference to the trainer object. Set by the trainer when the hook is registered.
    trainer: "torch_frame.Trainer" = None

    def before_train(self) -> None:
        """整体训练前调用"""
        pass

    def after_train(self) -> None:
        """全部训练结束后调用"""
        pass

    def before_epoch(self) -> None:
        """epoch前调用"""
        pass

    def after_epoch(self) -> None:
        """epoch结束后调用"""
        pass

    def before_iter(self) -> None:
        """iter前调用"""
        pass

    def after_iter(self) -> None:
        """iter结束后调用"""
        pass

    @property
    def checkpointable(self) -> bool:
        """A hook is checkpointable when it has :meth:`state_dict` method.
        Its state will be saved into checkpoint.
        """
        return callable(getattr(self, "state_dict", None))

    @property
    def class_name(self) -> str:
        """返回类名"""
        return self.__class__.__name__

    @property
    def metric_storage(self) -> "torch_frame.MetricStorage":
        return self.trainer.metric_storage

    def log(self, *args, **kwargs) -> None:
        self.trainer.log(*args, **kwargs)

    # belows are some helper functions that are often used in hook
    def every_n_epochs(self, n: int) -> bool:
        return (self.trainer.epoch + 1) % n == 0 if n > 0 else False

    def every_n_iters(self, n: int) -> bool:
        return (self.trainer.iter + 1) % n == 0 if n > 0 else False

    def every_n_inner_iters(self, n: int) -> bool:
        return (self.trainer.inner_iter + 1) % n == 0 if n > 0 else False

    def is_last_epoch(self) -> bool:
        return self.trainer.epoch == self.trainer.max_epochs - 1

    def is_last_iter(self) -> bool:
        return self.trainer.iter == self.trainer.max_iters - 1

    def is_last_inner_iter(self) -> bool:
        return self.trainer.inner_iter == self.trainer.epoch_len - 1



from typing import Optional, List
from torch.optim.lr_scheduler import _LRScheduler


class LRWarmupScheduler(_LRScheduler):
    """
    支持warmup的LR调节器。封装了LR scheduler, 可支持warmup, 每到iter结束后便调用。与pytorch不同的是, 该类的调用单位是iter
    而非epoch

    .. code-block:: python
        :emphasize-lines: 15-18

        lr_scheduler = LRWarmupScheduler(
            StepLR(optimizer, step_size=10, gamma=0.1),
            epoch_len=9999,  # 9999 iterations per epoch
            warmup_method="linear",
            warmup_iters=1000
            warmup_factor=0.001
        )
        for epoch in range(start_epoch, max_epochs):
            for batch in data_loader:
                output = model(batch)
                loss = criterion(output, target)
                optimizer.zero_grad()
                loss.backward()
                optimizer.step()
                # should be here
                lr_scheduler.step()
            # not here
            # lr_scheduler.step()
    """

    def __init__(
        self,
        scheduler: _LRScheduler,
        epoch_len: int,
        warmup_method: Optional[str] = None,
        warmup_iters: int = 1000,
        warmup_factor: float = 0.001,
        last_epoch: int = -1,
    ):
        """

        Parameters
        ----------
        scheduler : torch.optim.lr_scheduler._LRScheduler. 一个pytorch的标准调节器
        epoch_len : int. 一个epoch的长度, 用于在epoch结束时调用LR调节器
        warmup_method : str, default None.
            warmup的类型
            * constant, 常量
            * linear, 线性
            * exp, 指数
            * None, 不使用warmup
        warmup_iters : int, default 1000. warmup的总共iter数
        warmup_factor : float, default 0.001. warmup最初的学习率 = warmup_factor * 初始学习率
        last_epoch : int, default -1.
        """
        self.scheduler = scheduler
        self.epoch_len = epoch_len
        self.warmup_method = warmup_method
        self.warmup_iters = warmup_iters
        self.warmup_factor = warmup_factor

        if self._enable_warmup():
            assert warmup_method in ["constant", "linear", "exp"], (
                f"'{warmup_method}' is not a supported type for warmup, "
                "valid types are 'constant', 'linear' or 'exp'"
            )
            assert callable(
                getattr(scheduler, "_get_closed_form_lr", None)
            ), "`scheduler` must implement `_get_closed_form_lr()` method"
            assert warmup_iters > 0, "'warmup_iters' must be a positive integer"
            assert 0 < warmup_factor <= 1.0, "'warmup_ratio' must be in range (0,1]"

        # expected lr if no warming up is performed
        self.regular_lrs = scheduler.get_last_lr()

        super().__init__(scheduler.optimizer, last_epoch)

    def _enable_warmup(self) -> bool:
        return self.warmup_method is not None

    def _reach_epoch_end(self) -> bool:
        return self.last_epoch and self.last_epoch % self.epoch_len == 0

    def _get_warmup_factor(self) -> float:
        # `self.last_epoch` should be understood as `self.last_iter`
        if not self._enable_warmup() or self.last_epoch >= self.warmup_iters:
            return 1.0

        alpha = self.last_epoch / self.warmup_iters
        if self.warmup_method == "constant":
            return self.warmup_factor
        elif self.warmup_method == "linear":
            return self.warmup_factor * (1 - alpha) + alpha
        else:
            return self.warmup_factor ** (1 - alpha)

    def get_lr(self) -> List[float]:
        warmup_factor = self._get_warmup_factor()
        if self._reach_epoch_end():
            # `self.scheduler.last_epoch` is really the last epoch
            self.scheduler.last_epoch += 1
            self.regular_lrs = self.scheduler._get_closed_form_lr()
        return [warmup_factor * lr for lr in self.regular_lrs]

    def state_dict(self):
        state = {
            key: value
            for key, value in self.__dict__.items()
            if key != "optimizer" and key != "scheduler"
        }
        state["scheduler_state_dict"] = self.scheduler.state_dict()
        return state

    def load_state_dict(self, state_dict):
        self.scheduler.load_state_dict(state_dict.pop("scheduler_state_dict"))
        self.__dict__.update(state_dict)



import os
import setuptools

# 如果readme文件中有中文，那么这里要指定encoding='utf-8'，否则会出现编码错误
with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as readme:
    README = readme.read()

# 允许setup.py在任何路径下执行
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setuptools.setup(
    name="torch-frame",  # 库名, 需要在pypi中唯一
    version="1.7.4",  # 版本号
    author="Darkn Lxs",  # 作者
    author_email="1187220556@qq.com",  # 作看都将（方便使用索类现问图后成我我们）
    description="用于深度学习快速实现代码的框架",  # 简介
    long_description="见readme",  # 详细描述（一般会写在README.md中）
    long_description_content_type="text/markdown",  # README.md中描述的语法（一般为markdown)
    url="https://github.com/darknli/Pytorch-Frame/tree/main/torch_frame",  # 库/项目主页，放该项目的远程库地址即可
    packages=setuptools.find_packages(),  # 默认值即可，这个是方便以后我们给库拓展新功能的
    classifiers=[  # 指定该库依赖的Python版本、license、操作系统之类的
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=[  # 该库需要的依前库
        "termcolor",
        "numpy>=1.17",
        "opencv-python",
        "tabulate",
        "torch",
        "transformers>=4.25.1",
        "accelerate>=0.16.0",
        "diffusers",
        "pyyaml",
        "tqdm"
    ],
    python_requires='>=3.6',
)

# python setup.py sdist bdist_wheel


from typing import Union
import numpy as np
import cv2
import random
import math
from ..tools.object_detection import box_candidates

__all__ = [
    "random_perspective",
    "mixup_boxes",
    "mosaic"
]


def random_perspective(img: np.ndarray, targets: Union[np.ndarray, list, tuple], angle: int = 10, translate: int = 0.1,
                       scale: tuple = (0.7, 1.5), shear: float = 2.0, border: tuple = (0, 0),
                       fill_value: tuple = (128, 128, 128)):
    """
    对img做随机仿射变换，并且对应的boxes也做相应变化

    Parameters
    ----------
    img : np.ndarray. 原始图像.
    targets : np.ndarray, list or tuple. 目标box列表, 格式为[[x1, y1, x2, y2], ..., [x1, y1, x2, y2]].
    angle : float or int. 旋转图像的最大度数（含正负）.
    translate : int, default 0.1. 随机沿着宽高平移的百分比, 比如0.1代表分别向x和y方向平移宽和高的最大0.1距离.
    scale : tuple, default (0.7, 1.5). 缩放的范围, 第一个值代表最小界限，第二个值代表最大界限.
    shear : float, default 2. 仿射变换的不规则形变的范围, 如果是0，那么就不做仿射变换.
    border : tuple, default (0, 0). 向外padding的个数， 第一个参数是高height，第二个是宽width.
    fill_value: tuple, default (128, 128, 128). padding填充的颜色

    Returns
    -------
    img : np.ndarray. 做过仿射变换的图片
    targets : np.ndarray. 经过仿射变换后做过过滤的boxes
    """
    if isinstance(targets, tuple) or isinstance(targets, tuple):
        targets = np.array(targets, dtype=np.float)

    height = img.shape[0] + border[0] * 2  # shape(h,w,c)
    width = img.shape[1] + border[1] * 2

    padding_image = np.full((height, width, 3), fill_value, dtype=np.uint8)
    padding_image[border[0]:-border[0], border[1]:-border[1]] = img
    img = padding_image

    # Center
    C = np.eye(3)
    C[0, 2] = -img.shape[1] / 2  # x translation (pixels)
    C[1, 2] = -img.shape[0] / 2  # y translation (pixels)

    # Rotation and Scale
    R = np.eye(3)
    a = random.uniform(-angle, angle)
    # a += random.choice([-180, -90, 0, 90])  # add 90deg rotations to small rotations
    s = random.uniform(scale[0], scale[1])
    # s = 2 ** random.uniform(-scale, scale)
    R[:2] = cv2.getRotationMatrix2D(angle=a, center=(0, 0), scale=s)

    # Shear
    S = np.eye(3)
    S[0, 1] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # x shear (deg)
    S[1, 0] = math.tan(random.uniform(-shear, shear) * math.pi / 180)  # y shear (deg)
    perspective = abs(shear) < 1e-3

    # Translation
    T = np.eye(3)
    T[0, 2] = (
            random.uniform(0.5 - translate, 0.5 + translate) * width
    )  # x translation (pixels)
    T[1, 2] = (
            random.uniform(0.5 - translate, 0.5 + translate) * height
    )  # y translation (pixels)

    # Combined rotation matrix
    M = T @ S @ R @ C  # order of operations (right to left) is IMPORTANT, @ means matrix multiplication

    ###########################
    # For Aug out of Mosaic
    # s = 1.
    # M = np.eye(3)
    ###########################

    if (border[0] != 0) or (border[1] != 0) or (M != np.eye(3)).any():  # image changed
        if perspective:
            img = cv2.warpPerspective(img, M, dsize=(width, height), borderValue=fill_value)
        else:  # affine
            img = cv2.warpAffine(img, M[:2], dsize=(width, height), borderValue=fill_value)

    # Transform label coordinates
    n = len(targets)
    if n:
        # warp points
        xy = np.ones((n * 4, 3))
        xy[:, :2] = targets[:, [0, 1, 2, 3, 0, 3, 2, 1]].reshape(
            n * 4, 2
        )  # x1y1, x2y2, x1y2, x2y1
        xy = xy @ M.T  # transform
        if perspective:
            xy = (xy[:, :2] / xy[:, 2:3]).reshape(n, 8)  # rescale
        else:  # affine
            xy = xy[:, :2].reshape(n, 8)

        # create new boxes
        x = xy[:, [0, 2, 4, 6]]
        y = xy[:, [1, 3, 5, 7]]
        xy = np.concatenate((x.min(1), y.min(1), x.max(1), y.max(1))).reshape(4, n).T

        # clip boxes
        xy[:, [0, 2]] = xy[:, [0, 2]].clip(0, width)
        xy[:, [1, 3]] = xy[:, [1, 3]].clip(0, height)

        # filter candidates
        i = box_candidates(box1=targets[:, :4].T * s, box2=xy.T)
        targets = targets[i]
        targets[:, :4] = xy[i]

    return img, targets


def mixup_boxes(img1: np.ndarray, boxes1: np.ndarray, img2: np.ndarray, boxes2: np.ndarray,
                mixup_scale: float = 0.1, wh_thr: int = 0, ar_thr: int = 100, area_thr: float = 0):
    """
    mixup的目标检测版

    Parameters
    ----------
    img1 : np.ndarray. 图片1
    boxes1 : np.ndarray. 图片1对应的boxes
    img2 : np.ndarray. 图片2
    boxes2 : np.ndarray. 图片2对应的boxes
    mixup_scale: float, default 0.1. 做alpha融合的alpha波动概率，比如0.1的话实际alpha值会在0.4~0.6均匀采样
    wh_thr : int, default 0. boxes宽和高的最小阈值，低于它的box会被删除
    ar_thr ： int, default 100. 宽高比，低于它的box会被删除
    area_thr : float, default 0. 旧vs新box面积比例，低于它的box被删除

    Returns
    -------
    dst_img : np.ndarray
    dst_boxes : np.ndarray
    """
    def translate(image, boxes, h, w, alpha):
        y_offset = random.randint(0, dst_img.shape[0] - h)
        x_offset = random.randint(0, dst_img.shape[1] - w)
        dst_img[y_offset: h + y_offset, x_offset: w + x_offset] += image * alpha
        if len(boxes) > 0:
            new_boxes = boxes.copy()
            new_boxes[:, 1::2] = np.clip(new_boxes[:, 1::2] + x_offset, 0, dst_w)
            new_boxes[:, 2::2] = np.clip(new_boxes[:, 2::2] + y_offset, 0, dst_h)
            mask = box_candidates(boxes, new_boxes[:, 1:], wh_thr, ar_thr, area_thr)
            boxes = new_boxes[mask]
        else:
            boxes = np.zeros((0, 5))
        return dst_img, boxes

    h1, w1 = img1.shape[:2]
    h2, w2 = img2.shape[:2]
    dst_img = np.zeros((max(h1, h2), max(w1, w2), 3), dtype=np.float)
    dst_h, dst_w = dst_img.shape[:2]
    alpha = random.uniform(0.5 - mixup_scale, 0.5 + mixup_scale)
    img1 = img1.astype(np.float32)
    dst_img, dst_boxes1 = translate(img1.astype(float), boxes1, h1, w1, alpha)
    dst_img, dst_boxes2 = translate(img2.astype(float), boxes2, h2, w2, 1 - alpha)
    dst_boxes = np.vstack([dst_boxes1, dst_boxes2])
    return dst_img.astype(np.uint8), dst_boxes


def mosaic(data: list, ouput_dim: Union[list, tuple], fill_value=114):
    """
    mosaic拼接，用于数据增强，对于小目标检测有效

    Parameters
    ----------
    data : List[tuple]. list包含(image, bboxes)tuple，长度是4
    ouput_dim : Union[list, tuple]. 输出是[height, width]
    fill_value : Union[int, tuple].

    Returns
    -------
    mosaic_img : np.ndarray. 拼接后的图片
    mosaic_bboxes : Union[list, np.ndarray]. 拼接后的目标框，为空时是list类型
    """

    def get_mosaic_coordinate(mosaic_index, xc, yc, w, h):
        # index0 to top left part of image
        if mosaic_index == 0:
            x1, y1, x2, y2 = max(xc - w, 0), max(yc - h, 0), xc, yc
            small_coord = w - (x2 - x1), h - (y2 - y1), w, h
        # index1 to top right part of image
        elif mosaic_index == 1:
            x1, y1, x2, y2 = xc, max(yc - h, 0), min(xc + w, ouput_w * 2), yc
            small_coord = 0, h - (y2 - y1), min(w, x2 - x1), h
        # index2 to bottom left part of image
        elif mosaic_index == 2:
            x1, y1, x2, y2 = max(xc - w, 0), yc, xc, min(ouput_h * 2, yc + h)
            small_coord = w - (x2 - x1), 0, w, min(y2 - y1, h)
        # index2 to bottom right part of image
        elif mosaic_index == 3:
            x1, y1, x2, y2 = xc, yc, min(xc + w, ouput_w * 2), min(ouput_h * 2, yc + h)  # noqa
            small_coord = 0, 0, min(w, x2 - x1), min(y2 - y1, h)
        return (x1, y1, x2, y2), small_coord

    assert len(data) == 4, "`data`长度应该是4"
    assert len(ouput_dim) == 2, "`data`长度应该是2"

    mosaic_bboxes = []
    ouput_h, ouput_w = ouput_dim[0], ouput_dim[1]

    # yc, xc = s, s  # mosaic center x, y
    yc = int(random.uniform(0.5 * ouput_h, 1.5 * ouput_h))
    xc = int(random.uniform(0.5 * ouput_w, 1.5 * ouput_w))
    mosaic_img = np.full((ouput_h * 2, ouput_w * 2, 3), fill_value, dtype=np.uint8)

    for i_mosaic, (img, bboxes) in enumerate(data):
        h0, w0 = img.shape[:2]  # orig hw
        scale = min(1. * ouput_h / h0, 1. * ouput_w / w0)
        img = cv2.resize(
            img, (int(w0 * scale), int(h0 * scale)), interpolation=cv2.INTER_LINEAR
        )
        h, w = img.shape[:2]

        # suffix l means large image, while s means small image in mosaic aug.
        (l_x1, l_y1, l_x2, l_y2), (s_x1, s_y1, s_x2, s_y2) = get_mosaic_coordinate(i_mosaic, xc, yc, w, h)

        mosaic_img[l_y1:l_y2, l_x1:l_x2] = img[s_y1:s_y2, s_x1:s_x2]
        padw, padh = l_x1 - s_x1, l_y1 - s_y1

        # Normalized xywh to pixel xyxy format
        if bboxes.size > 0:
            scale_bboxes = bboxes.copy()
            scale_bboxes[:, 1] = scale * bboxes[:, 1] + padw
            scale_bboxes[:, 2] = scale * bboxes[:, 2] + padh
            scale_bboxes[:, 3] = scale * bboxes[:, 3] + padw
            scale_bboxes[:, 4] = scale * bboxes[:, 4] + padh
        else:
            scale_bboxes = np.zeros((0, 5))
        mosaic_bboxes.append(scale_bboxes)

    if len(mosaic_bboxes) > 0:
        mosaic_bboxes = np.concatenate(mosaic_bboxes, 0)
        np.clip(mosaic_bboxes[:, 0], 0, 2 * ouput_w, out=mosaic_bboxes[:, 0])
        np.clip(mosaic_bboxes[:, 1], 0, 2 * ouput_h, out=mosaic_bboxes[:, 1])
        np.clip(mosaic_bboxes[:, 2], 0, 2 * ouput_w, out=mosaic_bboxes[:, 2])
        np.clip(mosaic_bboxes[:, 3], 0, 2 * ouput_h, out=mosaic_bboxes[:, 3])

    return mosaic_img, mosaic_bboxes
